import operator as op
import pytest

from sweetpea import *
from sweetpea._internal.server import build_cnf
from acceptance import shuffled_design_sample, path_to_cnf_files, reset_expected_solutions

# Basic setup
color_list = ["red", "blue"]
color = Factor("color", color_list)
text  = Factor("text",  color_list)
mix   = Factor("mix",  color_list)

# Congruent Factor
con_level  = DerivedLevel("con", WithinTrial(op.eq, [color, text]))
inc_level  = DerivedLevel("inc", WithinTrial(op.ne, [color, text]))
con_factor = Factor("congruent?", [con_level, inc_level])

# Repeated color Factor
repeated_color_factor = Factor("repeated color?", [
    DerivedLevel("yes", Transition(lambda colors: colors[0] == colors[-1], [color])),
    DerivedLevel("no",  Transition(lambda colors: colors[0] != colors[-1], [color]))
])

# Repeated text Factor
repeated_text_factor = Factor("repeated text?", [
    DerivedLevel("yes", Transition(lambda texts: texts[0] == texts[-1], [text])),
    DerivedLevel("no",  Transition(lambda texts: texts[0] != texts[-1], [text]))
])

@pytest.mark.parametrize('strategy', [RandomGen, IterateSATGen])
@pytest.mark.parametrize('add_transition', [False, True])
def test_correct_solutions_with_different_crossing_sizes(strategy, add_transition):
    mix   = Factor("mix",  ["cake", "concrete", "tape"])
    design = [color, mix, text] + ([repeated_color_factor] if add_transition else [])
    crossing = [[color, text] + ([repeated_color_factor] if add_transition else []), [text, mix]]
    constraints = []

    block  = MultiCrossBlock(design, crossing, constraints)
    experiments  = synthesize_trials(block, 10, RandomGen)

    start = 1 if add_transition else 0

    assert len(experiments[0]['color']) == (9 if add_transition else 6)

    for e in experiments:
        seen = {}
        if add_transition:
            for i in range(1, 9):
                key = (e['color'][i], e['text'][i], e['repeated color?'][i])
                assert key not in seen
                seen[key] = True
            for i in range(1, 9):
                key = (e['text'][i], e['mix'][i])
                if key in seen:
                    assert seen[key] == 1
                    seen[key] = 2
                else:
                    seen[key] = 1
        else:
            for i in range(6):
                key = (e['text'][i], e['mix'][i])
                assert key not in seen
                seen[key] = True
            for i in range(6):
                key = ('one', e['color'][i], e['text'][i])
                if key in seen:
                    assert seen[key] == 1
                    seen[key] = 2
                else:
                    seen[key] = 1

@pytest.mark.parametrize('strategy', [RandomGen, IterateSATGen])
@pytest.mark.parametrize('design', shuffled_design_sample([color, text, mix], 6))
def test_correct_solution_count(strategy, design):
    crossing = [[color, mix], [text, mix]]
    constraints = []

    block  = MultiCrossBlock(design, crossing, constraints)
    experiments  = synthesize_trials(block, 100, sampling_strategy=strategy)

    assert len(experiments) == 96


@pytest.mark.parametrize('strategy', [RandomGen, IterateSATGen])
@pytest.mark.parametrize('design', shuffled_design_sample([color, text, mix, con_factor], 6))
def test_correct_solution_count_with_congruence_factor_but_unconstrained(strategy, design):
    crossing = [[color, text], [text, mix]]
    constraints = []

    block  = MultiCrossBlock(design, crossing, constraints)
    experiments  = synthesize_trials(block, 100, sampling_strategy=strategy)

    assert len(experiments) == 96


@pytest.mark.parametrize('strategy', [RandomGen, IterateSATGen])
@pytest.mark.parametrize('design', shuffled_design_sample([color, text, mix, con_factor], 6))
def test_correct_solution_count_with_congruence_factor_and_constrained(strategy, design):
    crossing = [[color, text], [text, mix]]
    constraints = [AtMostKInARow(1, (con_factor, "con"))]

    block  = MultiCrossBlock(design, crossing, constraints)
    experiments  = synthesize_trials(block, 100, sampling_strategy=strategy)

    assert len(experiments) == 48


def test_correct_solution_count_with_congruence_factor_and_constrained_cnf(design=[color, text, mix, con_factor]):
    crossing = [[color, text], [text, mix]]
    constraints = [AtMostKInARow(1, (con_factor, "con"))]

    block  = MultiCrossBlock(design, crossing, constraints)
    cnf = build_cnf(block)

    if reset_expected_solutions:
        with open(path_to_cnf_files+'/test_correct_solution_count_with_congruence_factor_and_constrained.cnf', 'w') as f:
            f.write(cnf.as_unigen_string())
    with open(path_to_cnf_files+'/test_correct_solution_count_with_congruence_factor_and_constrained.cnf', 'r') as f:
        old_cnf = f.read()

    assert old_cnf == cnf.as_unigen_string()


@pytest.mark.parametrize('strategy', [RandomGen, IterateSATGen])
@pytest.mark.parametrize('design', shuffled_design_sample([color, text, mix, con_factor], 6))
def test_correct_solution_count_with_congruence_factor_and_constrained_exactly(strategy, design):
    crossing = [[color, text], [text, mix]]
    constraints = [ExactlyKInARow(2, con_factor)]

    block  = MultiCrossBlock(design, crossing, constraints)
    experiments  = synthesize_trials(block, 100, sampling_strategy=strategy)

    assert len(experiments) == 32


def test_correct_solution_count_with_congruence_factor_and_constrained_exactly_cnf(design=[color, text, mix, con_factor]):
    crossing = [[color, text], [text, mix]]
    constraints = [ExactlyKInARow(2, con_factor)]

    block  = MultiCrossBlock(design, crossing, constraints)
    cnf = build_cnf(block)

    if reset_expected_solutions:
        with open(path_to_cnf_files+'/test_correct_solution_count_with_congruence_factor_and_constrained_exactly.cnf', 'w') as f:
            f.write(cnf.as_unigen_string())
    with open(path_to_cnf_files+'/test_correct_solution_count_with_congruence_factor_and_constrained_exactly.cnf', 'r') as f:
        old_cnf = f.read()

    assert old_cnf == cnf.as_unigen_string()

@pytest.mark.parametrize('strategy', [RandomGen, IterateSATGen])
@pytest.mark.parametrize('design', shuffled_design_sample([color, text, mix, repeated_color_factor], 6))
def test_correct_solution_count_with_repeated_color_factor_but_unconstrained(strategy, design):
    crossing = [[color, text], [text, mix]]
    constraints = []

    block  = MultiCrossBlock(design, crossing, constraints)
    experiments  = synthesize_trials(block, 100, sampling_strategy=strategy)

    assert len(experiments) == 96

@pytest.mark.parametrize('strategy', [RandomGen, IterateSATGen])
@pytest.mark.parametrize('design', shuffled_design_sample([color, text, mix, repeated_color_factor], 6))
def test_correct_solution_count_with_repeated_color_factor_and_constrained(strategy, design):
    crossing = [[color, text], [mix, text]]
    constraints = [AtMostKInARow(1, (repeated_color_factor, "yes"))]

    block  = MultiCrossBlock(design, crossing, constraints)
    experiments  = synthesize_trials(block, 100, sampling_strategy=strategy)

    # With only two colors, there can never be two color repetitions anyways,
    # so the total should still be the same.
    assert len(experiments) == 96

@pytest.mark.parametrize('strategy', [RandomGen, IterateSATGen])
@pytest.mark.parametrize('design', shuffled_design_sample([color, text, mix, repeated_color_factor, repeated_text_factor], 6))
def test_correct_solution_count_with_repeated_color_and_text_factors_but_unconstrained(strategy, design):
    crossing = [[color, text], [mix, text]]
    constraints = []

    block  = MultiCrossBlock(design, crossing, constraints)
    experiments  = synthesize_trials(block, 100, sampling_strategy=strategy)

    assert len(experiments) == 96


@pytest.mark.parametrize('strategy', [RandomGen, IterateSATGen])
@pytest.mark.parametrize('design', shuffled_design_sample([color, text, mix, repeated_color_factor, repeated_text_factor], 6))
def test_correct_solution_count_with_repeated_color_and_text_factors_and_constrained(strategy, design):
    crossing = [[color, text], [mix, text]]
    constraints = [
        AtMostKInARow(1, (repeated_color_factor, "yes")),
        AtMostKInARow(1, (repeated_text_factor, "yes"))
    ]

    block  = MultiCrossBlock(design, crossing, constraints)
    experiments  = synthesize_trials(block, 100, sampling_strategy=strategy)

    assert len(experiments) == 96


@pytest.mark.parametrize('strategy', [RandomGen, IterateSATGen])
@pytest.mark.parametrize('design', shuffled_design_sample([color, text, mix, repeated_color_factor], 6))
def test_correct_solution_count_with_repeated_color_factor_and_no_repetition_allowed(strategy, design):
    crossing = [[color, text], [mix, text]]
    constraints = [Exclude(repeated_color_factor["yes"])]

    block  = MultiCrossBlock(design, crossing, constraints)
    experiments  = synthesize_trials(block, 100, sampling_strategy=strategy)

    assert len(experiments) == 32


def test_correct_solution_count_with_repeated_color_factor_and_no_repetition_allowed_cnf(design=[color, text, mix, repeated_color_factor]):
    crossing = [[color, text], [mix, text]]
    constraints = [Exclude(repeated_color_factor["yes"])]

    block  = MultiCrossBlock(design, crossing, constraints)
    cnf = build_cnf(block)

    if reset_expected_solutions:
        with open(path_to_cnf_files+'/test_correct_solution_count_with_repeated_color_factor_and_no_repetition_allowed.cnf', 'w') as f:
            f.write(cnf.as_unigen_string())
    with open(path_to_cnf_files+'/test_correct_solution_count_with_repeated_color_factor_and_no_repetition_allowed.cnf', 'r') as f:
        old_cnf = f.read()

    assert old_cnf == cnf.as_unigen_string()

@pytest.mark.parametrize('strategy', [RandomGen, IterateSATGen])
def test_works_with_much_smaller_first_crossing(strategy, design=[color, text, mix, repeated_color_factor]):
    a = Factor("a", ["a1", "a2"])
    b = Factor("b", ["b1", "b2"])
    c = Factor("c", ["c1", "c2"])
    design = [a, b, c]
    crossing = [[c], [a,b]]
    block=MultiCrossBlock(design,crossing,[])
    
    experiments = synthesize_trials(block, 1, strategy)
    assert len(experiments[0]["a"]) == 4
    
    for e in experiments:
        seen = {}
        for i in range(0, 4):
            key = e['c'][i]
            if key in seen:
                assert seen[key] == 1
                seen[key] = 2
            else:
                seen[key] = 1
        seen = {}
        for i in range(0, 4):
            key = (e['a'][i], e['b'][i])
            assert key not in seen
            seen[key] = True



"""This module provides the types and common functions that are used across the
various submodules of :mod:`sweetpea.core.generate`.
"""


from __future__ import annotations

from contextlib import contextmanager
from enum import Enum, auto
from io import TextIOWrapper
from pathlib import Path
from typing import Any, Dict, Iterator, List, NamedTuple, Optional
from uuid import uuid4 as generate_uuid

from ..cnf import CNF, Var


__all__ = [
    'AssertionType', 'GenerationRequest', 'SampleType', 'ProblemSpecification', 'Solution',
    'combine_and_save_cnf', 'combine_cnf_with_requests', 'save_cnf', 'temporary_cnf_file'
]


JSONDict = Dict[str, Any]


@contextmanager
def temporary_cnf_file(base_path: Path = Path('.'), suffix: str = ".cnf") -> Iterator[Path]:
    """Returns a :class:`pathlib.Path` to a new, local file in the directory of
    the given path with a ``.cnf`` suffix. When used as a context manager
    (recommended), the file will be deleted when it leaves the context scope.
    """
    cnf_file = base_path / Path(str(generate_uuid())).with_suffix(suffix)
    try:
        yield cnf_file
    finally:
        if cnf_file.exists():
            cnf_file.unlink()


class AssertionType(Enum):
    """The three supported variants of CNF assertion."""
    #: Assert k == n.
    EQ = auto()
    #: Assert k < n.
    LT = auto()
    #: Assert k > n.
    GT = auto()

    @staticmethod
    def from_json(s: str) -> AssertionType:
        """Converts a JSON string to an :class:`AssertionType`."""
        return AssertionType[s]


class GenerationRequest(NamedTuple):
    """A request to generate a CNF."""
    #: The variant of assertion to make.
    assertion_type: AssertionType
    #: The ``k`` value.
    k: int
    #: A list of variables to use in generation.
    boolean_values: List[Var]

    @staticmethod
    def from_json(data: JSONDict) -> GenerationRequest:
        """Converts a JSON object to a :class:`GenerationRequest`."""
        return GenerationRequest(
            assertion_type=AssertionType.from_json(data['equalityType']),
            k=data['k'],
            boolean_values=[Var(v) for v in data['booleanValues']])


class SampleType(Enum):
    """The supported methods of interacting with SweetPea core."""
    #: Uniform sampling of a CNF formula.
    Uniform       = auto()
    #: Non-uniform sampling of a CNF formula.
    NonUniform    = auto()
    #: Test whether a CNF formula is satisfiable.
    IsSatisfiable = auto()

    @staticmethod
    def from_json(s: str) -> SampleType:
        """Converts a JSON string to a :class:`SampleType`."""
        if s in ('BuildCNF', 'SampleUniform', 'Uniform'):
            return SampleType.Uniform
        elif s in ('SampleNonUniform', 'NonUniform'):
            return SampleType.NonUniform
        elif s in ('IsSAT', 'IsSatisfiable'):
            return SampleType.IsSatisfiable
        else:
            raise ValueError(f"Invalid value for SampleType: {s}")


class ProblemSpecification(NamedTuple):
    """A specification of a complete SweetPea problem to be solved."""
    #: The type of sample to produce.
    sample_type: SampleType
    #: The number of samples to take.
    sample_count: int
    #: The number of fresh variables in the input.
    # TODO: This should be removed eventually.
    fresh: int
    #: The length of the support set.
    support: int
    #: The CNF formula to sample.
    cnf: CNF
    #: A list of requests to make with regard to the CNF formula.
    requests: List[GenerationRequest]

    @staticmethod
    def from_json(data: JSONDict) -> ProblemSpecification:
        """Converts a JSON object to a :class:`ProblemSpecification`."""
        return ProblemSpecification(
            sample_type=SampleType.from_json(data['action']),
            sample_count=data['sampleCount'],
            fresh=data['fresh'],
            support=data['support'],
            cnf=CNF(data['cnfs']),
            requests=[GenerationRequest.from_json(gr) for gr in data['requests']])


class Solution(NamedTuple):
    """The result of a generation."""
    # TODO DOC
    assignment: List[int]
    # TODO DOC
    frequency: int


def combine_cnf_with_requests(initial_cnf: CNF,
                              fresh: int,
                              support: int,  # FIXME: Remove.
                              generation_requests: List[GenerationRequest]) -> CNF:
    """Combines a base :class:`CNF` with a new :class:`CNF` formed from the
    given :class:`GenerationRequests <.GenerationRequest>`.
    """
    fresh_cnf = CNF.from_fresh(fresh)
    for request in generation_requests:
        if request.assertion_type is AssertionType.EQ:
            fresh_cnf.assert_k_of_n(request.k, request.boolean_values)
        elif request.assertion_type is AssertionType.LT:
            fresh_cnf.assert_k_less_than_n(request.k, request.boolean_values)
        elif request.assertion_type is AssertionType.GT:
            fresh_cnf.assert_k_greater_than_n(request.k, request.boolean_values)
        else:
            raise ValueError(f"invalid assertion type: {request.assertion_type}")
    final_cnf = fresh_cnf + initial_cnf
    return final_cnf  # TODO: Does this still work right?


def save_cnf(filename: Path,
             cnf: CNF,
             fresh: Optional[int] = None,
             support: Optional[int] = None):
    """Writes a CNF formula to a file at the given path."""
    filename.write_text(cnf.as_unigen_string(support_set_length=support))


def combine_and_save_cnf(filename: Path,
                         initial_cnf: CNF,
                         fresh: int,
                         support: int,
                         generation_requests: List[GenerationRequest]):
    """Combines a base CNF formula with the augmentations specified by the
    :class:`list` of :class:`GenerationRequests <.GenerationRequest>`, merges
    those formulas, then saves the result to a file at the given path.
    """
    print("Encoding experiment constraints...")
    combined_cnf = combine_cnf_with_requests(initial_cnf, fresh, support, generation_requests)
    save_cnf(filename, combined_cnf, fresh, support)

def combine_and_save_opb(filename: Path,
                         cnf: CNF,
                         support: int,
                         generation_requests: List[GenerationRequest]):
    print("Encoding experiment constraints...")

    with open(filename, 'a') as opb_file:
        opb_file.write(cnf.as_opb_string())

        for request in generation_requests:
            if request.assertion_type is AssertionType.EQ:
                comparison = ' = ' + str(request.k)
            elif request.assertion_type is AssertionType.LT:
                comparison = ' <= ' + str(request.k - 1)
            elif request.assertion_type is AssertionType.GT:
                comparison = ' >= ' + str(request.k - 1)
            else:
                raise ValueError(f"invalid assertion type: {request.assertion_type}")
            opb_file.write('\n' + ' '.join(map(lambda x :  '+1 v' + str(x), request.boolean_values)) \
                                    + comparison + ' ; ')



import operator as op
import pytest

from sweetpea import *
from sweetpea._internal.sampling_strategy.guided import GuidedGen

# Basic setup
color_list = ["red", "blue"]
color = Factor("color", color_list)
text  = Factor("text",  color_list)

# Congruent Factor
con_level  = DerivedLevel("con", WithinTrial(op.eq, [color, text]))
inc_level  = DerivedLevel("inc", WithinTrial(op.ne, [color, text]))
con_factor = Factor("congruent?", [con_level, inc_level])

design      = [color, text, con_factor]
crossing    = [color, text]
constraints = []

block = CrossBlock(design, crossing, constraints)


def test_guided_sampling_works():
    trials = synthesize_trials(block, 5, sampling_strategy=GuidedGen)

    assert len(trials) == 5



from io import TextIOWrapper
from typing import List, Optional
from pathlib import Path

from ..cnf import CNF
from .utility import GenerationRequest, Solution, combine_and_save_opb, temporary_cnf_file


__all__ = ['sample_ilp']


def sample_ilp_iterate(count: int,
                    initial_cnf: CNF,
                    support: int,
                    generation_requests: List[GenerationRequest],
                    ) -> List[Solution]:
    with temporary_cnf_file(Path('.'), '.opb') as opb_file:
        combine_and_save_opb(opb_file, initial_cnf, support, generation_requests)
        print("Running Gurobi...")
        solutions = compute_solutions(opb_file, support, count)
        return [Solution(solution, 1) for solution in solutions]

def compute_solutions(filename: Path,
                      support: int,
                      count: int,
                      solutions: Optional[List[List[int]]] = None) -> List[List[int]]:
    try:
        import gurobipy
    except ImportError as e:
        raise Exception("To use a Gurobi-backed solver, please install the gurobipy package: pip install gurobipy\n")
    
    import gurobipy as gp
    from gurobipy import GRB

    with gp.Env(empty=True) as env:
        env.setParam('OutputFlag', 0)
        env.setParam('LogFile', '')
        env.start()
        while True:
            if solutions is None:
                solutions = []
            if count == 0:
                return solutions
            model = gp.read(filename.name, env)
            model.optimize()
            
            # TODO: Look into more error codes?
            # https://www.gurobi.com/documentation/9.5/refman/optimization_status_codes.html
            if model.Status != 2:
                return solutions

            solution = [int(v.varName.replace('v','')) \
                        if v.x == 1 \
                        else -int(v.varName.replace('v','')) \
                        for v in model.getVars()]
            solution = sorted(solution, key=abs)[:support]
            update_file(filename, solution)
            count -= 1
            solutions += [solution]

def update_file(filename: Path, solution: List[int]):
    with open(filename, 'a') as opb_file:
        false_count = len(list(v for v in solution if v < 0))
        opb_file.write('\n' + ' '.join(map(lambda x :  '-1 v' + str(abs(x)) \
                                                    if str(x)[0] == '-' \
                                                    else '+1 v' + str(x), solution)) \
                                + ' <= ' + str(len(solution) - 1 - false_count) + ' ;\n')



import operator as op
import pytest

from sweetpea import *
from acceptance import shuffled_design_sample, path_to_cnf_files, reset_expected_solutions

correct_response = Factor(name="correct_response", initial_levels=["H", "S"])
congruency = Factor(name="congruency", initial_levels=["congruent", "incongruent"])
design     = [correct_response, congruency]
crossing   = [correct_response, congruency]

@pytest.mark.parametrize('strategy', [IterateSATGen, RandomGen])
@pytest.mark.parametrize('trial_count', [4, 5, 6, 7, 8])
def test_stays_balanced(strategy, trial_count):
    crossing_size = 1
    for f in crossing:
        crossing_size *= len(f.levels)
    block  = Repeat(CrossBlock(design=design, crossing=crossing, constraints=[]),
                    constraints=[MinimumTrials(trials=trial_count)])
    experiments = synthesize_trials(block=block, samples=3, sampling_strategy = strategy)

    for exp in experiments:
        conds = list(zip(*exp.values()))
        for cond in conds:
            assert len(conds) == trial_count

        to_go = trial_count
        while to_go > 0:
            tabulation: Dict[List[str], bool] = dict()
            for cond in conds[:crossing_size]:
                assert not (cond in tabulation)
                tabulation[cond] = True
            to_go -= crossing_size
            cond = cond[crossing_size:]

@pytest.mark.parametrize('strategy', [IterateSATGen, RandomGen])
@pytest.mark.parametrize('trial_count', [4, 5, 6])
def test_leftovers_balanced_across_all_possible_solutions(strategy, trial_count):
    block  = Repeat(CrossBlock(design=design, crossing=crossing, constraints=[]),
                    constraints = [MinimumTrials(trials=trial_count)])
    experiments = synthesize_trials(block=block, samples=500, sampling_strategy = strategy)

    leftover = trial_count % 4

    def factorial(n):
        if n < 2:
            return 1
        else:
            return n * factorial(n - 1)

    assert len(experiments) == 24 * (24 // factorial(4 - leftover))

    # Make sure the solutions are actually distinct
    tabulation: Dict[List[str], bool] = dict()
    for e in experiments:
        key = tuple([(k, tuple(e[k])) for k in e])
        assert not key in tabulation
        tabulation[key] = True



"""This module provides some convenience functions to be used internally."""


from itertools import islice, tee, chain, repeat
from typing import Any, Tuple, List, Iterator, Iterable, Union, cast

from sweetpea._internal.primitive import Factor, DerivedLevel, SimpleLevel


def get_all_levels(design: List[Factor]) -> List[Tuple[Factor, Union[SimpleLevel, DerivedLevel]]]:
    return [(factor, cast(Union[SimpleLevel, DerivedLevel], level)) for factor in design for level in factor.levels]



import operator as op
import pytest

from sweetpea import Factor, DerivedLevel, WithinTrial, CrossBlock
from sweetpea._internal.server import build_cnf, is_cnf_still_sat
from sweetpea._internal.logic import And
from sweetpea._internal.constraint import Reify
from acceptance import path_to_cnf_files, reset_expected_solutions


# Basic setup
color_list = ["red", "blue"]
color = Factor("color", color_list)
text  = Factor("text",  color_list)

# Congruent Factor
con_level  = DerivedLevel("con", WithinTrial(op.eq, [color, text]))
inc_level  = DerivedLevel("inc", WithinTrial(op.ne, [color, text]))
con_factor = Factor("congruent?", [con_level, inc_level])

block = CrossBlock([color, text, con_factor], [color, text], [Reify(con_factor)])


def test_is_cnf_still_sat_should_respond_correctly():

    # Build the CNF on the server.
    cnf_result = build_cnf(block)

    if reset_expected_solutions:
        with open(path_to_cnf_files+'/test_is_cnf_still_sat_should_respond_correctly.cnf', 'w') as f:
            f.write(cnf_result.as_unigen_string())
    with open(path_to_cnf_files+'/test_is_cnf_still_sat_should_respond_correctly.cnf', 'r') as f:
        old_cnf = f.read()

    assert old_cnf == cnf_result.as_unigen_string()

    assert     is_cnf_still_sat(block, [And([1, 3])])

    assert not is_cnf_still_sat(block, [And([7, 8])])
    assert not is_cnf_still_sat(block, [And([1, 7, 13])])



import os
from typing import Tuple, List, Union
from itertools import repeat, permutations
from random import shuffle

from sweetpea._internal.primitive import Factor, SimpleLevel, DerivedLevel
from sweetpea._internal.constraint import  AtMostKInARow
from sweetpea._internal.level import get_all_levels
from sweetpea._internal.derivation_processor import DerivationProcessor

def __assert_atmostkinarow_pair(k: int, level: Tuple[Factor, Union[SimpleLevel, DerivedLevel]], experiments: List[dict]) -> None:
    sublist = list(repeat(level[1], k + 1))
    for e in experiments:
        assert sublist not in [e[level[0]][i:i+k+1] for i in range(len(e[level[0]]) - (k + 1))]

def __assert_atmostkinarow_factor(k: int, f: Factor, experiments: List[dict]) -> None:
    factor_name = f.name
    for level in f.levels:
        sublist = list(repeat(level.name, k + 1))
        for e in experiments:
            assert sublist not in [e[factor_name][i:i + k + 1] for i in range(len(e[factor_name]) - (k + 1))]

def assert_atmostkinarow(c: AtMostKInARow, experiments: List[dict]) -> None:
    if isinstance(c.level, Factor):
        __assert_atmostkinarow_factor(c.k, c.level, experiments)
    else:
        level_tuples = get_all_levels([c.level])
        for t in level_tuples:
            __assert_atmostkinarow_pair(c.k, t, experiments)

def assert_no_repetition(experiments: List[dict]) -> None:
    for seq in experiments:
        levels_lists = [levels for f, levels in seq.items()]
        transposed = list(map(list, zip(*levels_lists)))
        for t in transposed:
            assert transposed.count(t) == 1, "{} repeats in this trial sequence!".format(t)

def shuffled_design_sample(input, num):
    perms = list(permutations(input))
    shuffle(perms)
    return perms[:num]

# Set to True to reset recorded files that have specific CNF encodings,
# but leave as False to perform regression tests where the CNF should
# not change:
reset_expected_solutions = False

path_to_cnf_files = os.path.dirname(os.path.abspath(__file__)) + "/cnf_files"



import pytest

from sweetpea import *

@pytest.mark.parametrize('num_trials', [1, 2, 3, 4, 5, 6, 7, 8, 100, 1000])
def test_ok_to_ask_for_more_trials_than_solutions(num_trials):
    color      = Factor("color",  ["red", "green"])
    design       = [color]
    crossing     = [color]
    trial_constraint = MinimumTrials(num_trials)
    constraints = [trial_constraint]
    block        = Repeat(CrossBlock(design, crossing, [],
                                     require_complete_crossing=False),
                          constraints)
    experiments  = synthesize_trials(block, 1, sampling_strategy=RandomGen)

@pytest.mark.parametrize('strategy', [RandomGen, IterateSATGen])
def test_check_max_solutions(strategy):
    color      = Factor("color",  ["red", "green", "blue"])
    design       = [color]
    crossing     = [color]

    block        = CrossBlock(design, crossing, [MinimumTrials(3)],
                              require_complete_crossing=False)
    experiments  = synthesize_trials(block, 100, sampling_strategy=strategy)
    assert len(experiments) == 6

    block        = Repeat(CrossBlock(design, crossing, [],
                                     require_complete_crossing=False),
                          [MinimumTrials(3)])
    experiments  = synthesize_trials(block, 100, sampling_strategy=strategy)
    assert len(experiments) == 6

    block        = CrossBlock(design, crossing, [MinimumTrials(4)],
                              require_complete_crossing=False)
    experiments  = synthesize_trials(block, 100, sampling_strategy=strategy)
    assert len(experiments) == 54

    block        = Repeat(CrossBlock(design, crossing, [],
                                     require_complete_crossing=False),
                          [MinimumTrials(4)])
    experiments  = synthesize_trials(block, 100, sampling_strategy=strategy)
    assert len(experiments) == 18

    block        = CrossBlock(design, crossing, [MinimumTrials(5)],
                              require_complete_crossing=False)
    experiments  = synthesize_trials(block, 100, sampling_strategy=strategy)
    assert len(experiments) == 90

    block        = Repeat(CrossBlock(design, crossing, [],
                                     require_complete_crossing=False),
                          [MinimumTrials(5)])
    experiments  = synthesize_trials(block, 100, sampling_strategy=strategy)
    assert len(experiments) == 36

@pytest.mark.parametrize('strategy', [RandomGen, IterateSATGen])
def test_check_max_solutions_derived_factor(strategy):
    color = Factor("color", ['red', 'green'])
    size = Factor("size", ['small', 'med', 'large'])

    match = Factor(name="match", initial_levels=[
        DerivedLevel(name="high", window=WithinTrial(lambda a, b: b != 'large', [color, size])),
        DerivedLevel(name="low", window=WithinTrial(lambda a, b: b == 'large', [color, size]))
    ])

    block      = CrossBlock(design=[color, size, match], crossing=[color, match], constraints=[])
    experiments = synthesize_trials(block=block, samples=200, sampling_strategy=strategy)

    assert len(experiments) == 96



import operator as op

from sweetpea._internal.level import get_all_levels
from sweetpea._internal.iter import intersperse
from sweetpea._internal.primitive import Factor, DerivedLevel, Transition


color = Factor("color", ["red", "blue"])
text  = Factor("text",  ["red", "blue", "green"])

color_repeats_level   = DerivedLevel("yes", Transition(lambda colors: colors[0] == colors[-1], [color]))
color_no_repeat_level = DerivedLevel("no", Transition(lambda colors: colors[0] != colors[-1], [color]))
color_repeats_factor  = Factor("color repeats?", [color_repeats_level, color_no_repeat_level])

def test_get_all_external_level_names():
    assert get_all_levels([color, text]) == [(color, color.get_level('red')),
                                             (color, color.get_level('blue')),
                                             (text, text.get_level('red')),
                                             (text, text.get_level('blue')),
                                             (text, text.get_level('green'))]

    assert get_all_levels([color_repeats_factor]) == [(color_repeats_factor, color_repeats_level),
                                                      (color_repeats_factor, color_no_repeat_level)]

def test_intersperse():
    assert list(intersperse('', ['yes', 'no', 'yes'])) == \
        ['yes', '', 'no', '', 'yes']

    assert list(intersperse('', ['yes', 'no', 'yes'], 2)) == \
        ['yes', '', '', 'no', '', '', 'yes']

    assert list(intersperse('', ['yes', 'no', 'yes'], 0)) == \
        ['yes', 'no', 'yes']



import os
import numpy as np

from functools import reduce
from itertools import product, chain
from time import time
from typing import List, cast

from sweetpea._internal.block import Block
from sweetpea._internal.core import CNF, cnf_is_satisfiable
from sweetpea._internal.logic import And, cnf_to_json
from sweetpea._internal.sampling_strategy.base import Gen, SamplingResult
from sweetpea._internal.server import build_cnf


"""
This strategy gradually constructs samples in memory, with the aid of a SAT solver to
guide the choices it makes.

While sufficient in some cases, this strategy isn't guaranteed to produce uniform results
because trial selections early on can prune the remaining search space unevenly.
"""
class GuidedGen(Gen):

    @staticmethod
    def class_name():
        return 'GuidedGen'

    @staticmethod
    def sample(block: Block, sample_count: int) -> SamplingResult:

        samples = cast(List[dict], [])
        metrics = cast(dict, {
            'sample_metrics': []
        })

        overall_start = time()

        # Build the full CNF for this block
        cnf = build_cnf(block)

        metrics['solver_call_count'] = 0
        for _ in range(sample_count):
            sample_metrics = cast(dict, {})
            t_start = time()
            samples.append(GuidedGen.__generate_sample(block, cnf, sample_metrics))
            sample_metrics['time'] = time() - t_start
            metrics['sample_metrics'].append(sample_metrics)
            metrics['solver_call_count'] += sample_metrics['solver_call_count']

        metrics['time'] = time() - overall_start
        GuidedGen.__compute_additional_metrics(metrics)

        return SamplingResult(samples, metrics)


    @staticmethod
    def __generate_sample(block: Block, cnf: CNF, sample_metrics: dict) -> dict:
        sample_metrics['trials'] = []

        # Start a 'committed' list of CNFs
        committed = cast(List[And], [])

        for trial_number in range(block.trials_per_sample()):
            trial_start_time = time()

            trial_metrics = {
                't': trial_number + 1,
                'solver_calls': []
            }
            solver_calls = cast(List[dict], trial_metrics['solver_calls'])

            #  Get the variable list for this trial.
            variables = block.variable_list_for_trial(trial_number + 1)
            variables = list(filter(lambda i: i != [], variables))
            potential_trials = list(map(list, product(*variables)))

            trial_metrics['potential_trials'] = len(potential_trials)

            # Use env var to switch between filtering and not
            if GuidedGen.__prefilter_enabled():
                # Flatten the list
                flat_vars = list(chain(*variables))

                # Check SAT for each one
                unsat = []
                for v in flat_vars:
                    t_start = time()
                    full_cnf = cnf + CNF(cnf_to_json(committed)) + CNF(cnf_to_json([And([v])]))
                    allowed = cnf_is_satisfiable(full_cnf)
                    duration_seconds = time() - t_start
                    solver_calls.append({'time': duration_seconds, 'SAT': allowed})
                    if not allowed:
                        unsat.append(v)

                # TODO: Count filtering SAT calls separately?

                # Filter out any potential trials with those vars set
                filtered_pts = []
                for pt in potential_trials:
                    if any(uv in pt for uv in unsat):
                        continue
                    else:
                        filtered_pts.append(pt)

                # Record the number filterd out for metrics
                trial_metrics['prefiltered_out'] = len(potential_trials) - len(filtered_pts)
                potential_trials = filtered_pts

            allowed_trials = []
            for potential_trial in potential_trials:
                start_time = time()
                full_cnf = cnf + CNF(cnf_to_json(committed)) + CNF(cnf_to_json([And(potential_trial)]))
                allowed = cnf_is_satisfiable(full_cnf)
                duration_seconds = time() - start_time

                solver_calls.append({'time': duration_seconds, 'SAT': allowed})

                if allowed:
                    allowed_trials.append(potential_trial)

            trial_metrics['allowed_trials'] = len(allowed_trials)
            trial_metrics['solver_call_count'] = len(solver_calls)
            sample_metrics['trials'].append(trial_metrics)

            # Randomly sample a single trial from the uniform distribution of the allowed trials,
            # and commit that trial to the committed sequence.
            trial_idx = np.random.randint(0, len(allowed_trials))
            committed.append(And(allowed_trials[trial_idx]))

            trial_metrics['time'] = time() - trial_start_time

        # Aggregate the total solver calls
        sample_metrics['solver_call_count'] = 0
        for tm in sample_metrics['trials']:
            sample_metrics['solver_call_count'] += tm['solver_call_count']

        # Flatten the committed trials into a list of integers and decode it.
        solution = GuidedGen.__committed_to_solution(committed)
        return Gen.decode(block, solution)

    @staticmethod
    def __committed_to_solution(committed: List[And]) -> List[int]:
        return reduce(lambda sol, clause: sol + clause.input_list, committed, [])

    @staticmethod
    def __prefilter_enabled():
        return os.environ.get('SWEETPEA_GUIDED_PREFILTER_TRIALS') is not None

    @staticmethod
    def print_summary(result: SamplingResult) -> None:
        metrics = result.metrics
        print("Total Time: {} seconds".format(metrics['time']))
        print("Total SAT Solver Calls: {}".format(metrics['solver_call_count']))
        print("Mean SAT Time: {}".format(metrics['mean_sat_time']))
        print("Mean UNSAT Time: {}".format(metrics['mean_unsat_time']))

        if 'total_prefiltered_out' in metrics:
            print("Total Prefiltered Out: {}".format(metrics['total_prefiltered_out']))

    @staticmethod
    def __compute_additional_metrics(metrics: dict) -> None:
        total_sat = 0
        total_sat_time = 0
        total_unsat = 0
        total_unsat_time = 0

        for sample in metrics['sample_metrics']:
            for trial in sample['trials']:
                for sc in trial['solver_calls']:
                    if sc['SAT'] == True:
                        total_sat += 1
                        total_sat_time += sc['time']
                    elif sc['SAT'] == False:
                        total_unsat += 1
                        total_unsat_time += sc['time']

        metrics['mean_sat_time'] = total_sat_time / total_sat
        metrics['mean_unsat_time'] = total_unsat_time / total_unsat

        if GuidedGen.__prefilter_enabled():
            total_prefiltered_out = 0
            for sample in metrics['sample_metrics']:
                for trial in sample['trials']:
                    total_prefiltered_out += trial['prefiltered_out']
            metrics['total_prefiltered_out'] = total_prefiltered_out


"""
Generates a static HTML file that will render a flamegraph showing the time breakdown for a given sampling.
"""
class Flamegraph():
    GRAPH_FILE_TEMPLATE = '''
<head>
  <link rel="stylesheet" type="text/css" href="https://cdn.jsdelivr.net/gh/spiermar/d3-flame-graph@2.0.6/dist/d3-flamegraph.css">
</head>
<body>
  <div id="chart"></div>
  <script type="text/javascript" src="https://d3js.org/d3.v4.min.js"></script>
  <script type="text/javascript" src="https://cdnjs.cloudflare.com/ajax/libs/d3-tip/0.9.1/d3-tip.min.js"></script>
  <script type="text/javascript" src="https://cdn.jsdelivr.net/gh/spiermar/d3-flame-graph@2.0.6/dist/d3-flamegraph.min.js"></script>
  <script type="text/javascript">
  graph_data = {}
  var flamegraph = d3.flamegraph().width(960);
  d3.select("#chart").datum(graph_data).call(flamegraph);
  <!-- TODO: Render SAT/UNSAT calls as Green/Red -->
  <!-- TODO: Replace 'samples' with 'seconds' in tooltip -->
  </script>
</body>
'''

    @staticmethod
    def generate(filename: str, sampling_result: SamplingResult) -> None:
        graph_data = Flamegraph.__convert_metrics_to_graph_data(sampling_result.metrics)
        with open(filename, 'w') as f:
            f.write(Flamegraph.GRAPH_FILE_TEMPLATE.format(graph_data))

    @staticmethod
    def __convert_metrics_to_graph_data(metrics: dict) -> dict:
        children = []
        for sample_number in range(len(metrics['sample_metrics'])):
            children.append(Flamegraph.__convert_sample_to_graph_data(sample_number, metrics['sample_metrics'][sample_number]))

        return {
            'name': 'All Samples',
            'value': metrics['time'],
            'children': children
        }

    @staticmethod
    def __convert_sample_to_graph_data(sample_number: int, sample_metrics: dict) -> dict:
        children = []
        for trial_number in range(len(sample_metrics['trials'])):
            children.append(Flamegraph.__convert_trial_to_graph_data(trial_number, sample_metrics['trials'][trial_number]))

        return {
            'name': 'Sample #{}'.format(sample_number + 1),
            'value': sample_metrics['time'],
            'children': children
        }

    @staticmethod
    def __convert_trial_to_graph_data(trial_number: int, trial_metrics: dict) -> dict:
        children = []
        for solver_number in range(len(trial_metrics['solver_calls'])):
            solver_call = trial_metrics['solver_calls'][solver_number]
            children.append({
                'name': 'Solver Call #{}'.format(solver_number + 1),
                'value': solver_call['time'],
                'SAT': str(solver_call['SAT']).lower()
            })

        return {
            'name': 'Trial #{}'.format(trial_number + 1),
            'value': trial_metrics['time'],
            'children': children
        }



"""This module provides the core data types used in SweetPea Core.

Fundamentally, SweetPea deals with Boolean logic formulas written in
*conjunctive normal form (CNF)*. `Wikipedia says
<https://en.wikipedia.org/wiki/Conjunctive_normal_form>`_:

    a formula is in **conjunctive normal form (CNF)** [...] if it is a
    conjunction of one or more clauses, where a clause is a disjunction of
    literals; otherwise put, it is a **product of sums** or **an AND of ORs**.

In SweetPea, we refer to the "literals" as :class:`Vars <.Var>`. These are
combined into :class:`Clauses <.Clause>`, which are then combined into
:class:`CNFs <.CNF>`. Most SweetPea functionality revolves around building and
manipulating :class:`CNFs <.CNF>`.
"""

# Allow type annotations to refer to not-yet-declared types.
# flake8 doesn't know about this so we have to #noqa it.
from __future__ import annotations  # noqa

import math

from itertools import chain
from typing import Iterable, List, Optional, Sequence, Tuple, Union, cast

from .binary import BinaryNumber, int_to_binary
from .simple_sequence import SimpleSequence


__all__ = ['Var', 'Clause', 'CNF']

class Var:
    """A variable for use in a CNF formula.

    This is essentially a wrapper for the builtin :class:`int` type, but with a
    few advantages:

      * It is semantically distinct from an :class:`int`, so reading code is
        more sensible.
      * Type aliases for :class:`int` can be obnoxious to deal with because you
        may want to use :class:`int`-supported operations (like addition,
        negation, etc), but these are not cross-type compatible (e.g.,
        ``Var(3) + 2`` produces an error in mypy).
      * We can implement custom behavior as a sort of micro-DSL without much
        overhead or issue.
    """

    def __init__(self, value: int):
        self._val: int
        if isinstance(value, Var):
            self._val = value._val
        elif isinstance(value, int):
            if value == 0:
                raise ValueError(f"Var values must be non-zero integers; got {value}")
            self._val = value
        else:
            raise TypeError(f"expected 'int'; got '{type(value).__name__}'")

    @property
    def value(self) -> int:
        """The integer value of this variable."""
        return self._val

    def __repr__(self) -> str:
        return f"Var({self._val})"

    def __str__(self) -> str:
        return str(self._val)

    def __int__(self) -> int:
        return self._val

    def __hash__(self) -> int:
        return hash(self._val)

    def __eq__(self, other) -> bool:
        if isinstance(other, Var):
            return int(self) == int(other)
        return NotImplemented

    def __lt__(self, other: Var) -> bool:
        if isinstance(other, Var):
            return int(self) < int(other)
        return NotImplemented

    def __invert__(self) -> Var:
        """Logical NOT."""
        return Var(-self._val)

    def __or__(self, other: Var) -> Clause:
        """Logical OR."""
        if isinstance(other, Var):
            return Clause(self, other)
        return NotImplemented

    def __and__(self, other: Var) -> CNF:
        """Logical AND."""
        if isinstance(other, Var):
            return CNF(Clause(self), Clause(other))
        return NotImplemented

    def __xor__(self, other: Var) -> CNF:
        """Logical XOR."""
        if isinstance(other, Var):
            return CNF([[self, other], [~self, ~other]])
        return NotImplemented

    def __mod__(self, other: Var) -> CNF:
        """Logical XNOR."""
        # NOTE: This method is used to implement logical XNOR instead of
        #       modulo. If only Python allowed custom operators.
        if isinstance(other, Var):
            return CNF([[self, ~other], [~self, other]])
        return NotImplemented

    def __abs__(self) -> Var:
        return Var(abs(int(self)))


class Clause(SimpleSequence[Var]):
    """A sequence of :class:`Vars <.Var>`. A :class:`Clause` indicates a
    logical disjunction between the contained variables. For example,
    ``Clause(Var(3), Var(7))`` encodes (3 ∨ 7).

    A :class:`Clause` can also be instantiated with a :class:`list` of
    :class:`Vars <.Var>`. This mode of instantiation will also accept raw
    :class:`ints <int>` in addition to instances of :class:`Var`. For example,
    ``Clause([1, -2, 3])`` corresponds to the formula (1 ∨ ¬2 ∨ 3).
    """

    @classmethod
    def _get_element_type(cls):
        return Var

    def __str__(self) -> str:
        return ' '.join(str(var) for var in self)

    def __add__(self, other: Union[Clause, Var]) -> Clause:
        """Logical OR. This alias exists due to the :class:`list`-like
        interface of :class:`Clauses <.Clause>`.
        """
        if isinstance(other, Clause):
            return Clause(*self, *other)
        if isinstance(other, Var):
            return Clause(*self, other)
        return NotImplemented

    def __radd__(self, other: Var) -> Clause:
        if isinstance(other, Var):
            return Clause(other, *self)
        return NotImplemented

    def __and__(self, other: Union[Clause, Var]) -> CNF:
        """Logical AND."""
        if isinstance(other, Clause):
            return CNF(self, other)
        if isinstance(other, Var):
            return CNF(self, Clause(other))
        return NotImplemented

    def __rand__(self, other: Var) -> CNF:
        if isinstance(other, Var):
            return CNF(Clause(other), self)
        return NotImplemented

    def __or__(self, other: Union[Clause, Var]):
        """Logical OR."""
        return self + other

    def __ror__(self, other: Var):
        return other + self


class CNF(SimpleSequence[Clause]):
    """A conjunction of disjunction :class:`Clauses <.Clause>`. For example,
    ``CNF(Clause(Var(3), Var(7)), Clause(Var(1), Var(13)))`` corresponds to the
    CNF formula ((3 ∨ 7) ∧ (1 ∨ 13)).

    :class:`CNFs <.CNF>` can also be instantiated with a :class:`list` of
    :class:`lists <list>` of :class:`Vars <.Var>`, where each inner
    :class:`list` represents a :class:`Clause` and the outer :class:`list`
    represents the CNF formula itself. Similar to :class:`Clause`, this mode of
    instantiation will also accept raw :class:`ints <int>` in addition to
    instances of :class:`Var`. For example, ``CNF([[1, 2, -3], [-2, 7, 1]])``
    corresponds to the CNF formula ((1 ∨ 2 ∨ ¬3) ∧ (¬2 ∨ 7 ∨ 1)).
    """

    ########################################
    ##
    ## Static Methods
    ##

    ## This method is for initializing a CNF from a given number of fresh vars.
    @staticmethod
    def from_fresh(fresh: int) -> CNF:
        """Returns an empty :class:`CNF` with the given number of fresh
        :class:`Vars <.Var>` already allocated.

        .. note::

            This method probably shouldn't exist! The number of fresh variables
            should be deduced from the formulas themselves. This exists for
            legacy compatibility and should eventually be removed.
        """
        cnf = CNF()
        cnf._num_vars = fresh
        return cnf

    ## These are used for creating CNF formulas by combining two variables in a
    ## particular way.

    @staticmethod
    def and_vars(a: Union[int, Var], b: Union[int, Var]) -> CNF:
        """Returns a CNF formula encoding (a ∧ b)."""
        if not isinstance(a, Var):
            a = Var(a)
        if not isinstance(b, Var):
            b = Var(b)
        return a & b

    @staticmethod
    def or_vars(a: Union[int, Var], b: Union[int, Var]) -> CNF:
        """Returns a CNF formula encoding (a ∨ b)."""
        if not isinstance(a, Var):
            a = Var(a)
        if not isinstance(b, Var):
            b = Var(b)
        return CNF(a | b)

    @staticmethod
    def xor_vars(a: Union[int, Var], b: Union[int, Var]) -> CNF:
        """Returns a CNF formula encoding (a ⊕ b) as ((a ∨ b) ∧ (¬a ∨ ¬b))."""
        if not isinstance(a, Var):
            a = Var(a)
        if not isinstance(b, Var):
            b = Var(b)
        return a ^ b

    @staticmethod
    def xnor_vars(a: Union[int, Var], b: Union[int, Var]) -> CNF:
        """Returns a CNF formula encoding (a ⊙ b) as ((a ∨ ¬b) ∧ (¬a ∨ b)).

        .. note::

            (a ⊙ b) is logically equivalent to (a ⇔ b), so this function
            replaces the original Haskell version's |doubleImpliesLink|_.

        .. |doubleImpliesLink| replace:: ``doubleImplies`` function
        .. _doubleImpliesLink:
            https://github.com/sweetpea-org/sweetpea-core/blob/20493dd1ffc8f5a53bd4004286e6a64298a16ce3/src/DataStructures.hs#L122-L124
        """
        if not isinstance(a, Var):
            a = Var(a)
        if not isinstance(b, Var):
            b = Var(b)
        return a % b

    @staticmethod
    def distribute(v: Var, cnf: CNF) -> CNF:
        """Distributes the given :class:`Var` across each :class:`Clause` of
        the given CNF formula, producing a new :class`CNF`.
        """
        return v ** cnf

    ########################################
    ##
    ## Class Configuration/Initialization
    ##

    @classmethod
    def _get_element_type(cls):
        return Clause

    _num_vars: int

    def __init__(self, *values):
        super().__init__(*values)
        self._num_vars = len({abs(var) for clause in self._vals for var in clause})

    ########################################
    ##
    ## String Rendering
    ##

    def __str__(self) -> str:
        return ''.join(str(clause) + ' 0\n' for clause in reversed(self._vals))

    def as_opb_string(self) -> str:
        def count_false_var(clause : Clause):
            return len(list(v for v in clause if str(v)[0] == '-'))

        return '\n'.join(' '.join(map(lambda v : '-1 v' + str(v)[1:]
                                    if str(v)[0] == '-'
                                    else '+1 v' + str(v), clause)) \
                                + ' >= ' + str(-count_false_var(clause) + 1) + ' ;'
                                for clause in reversed(self._vals))

    def as_dimacs_string(self, fresh_variable_count: Optional[int] = None) -> str:
        """Represents the :class:`CNF` as a string in the DIMACS format.

        The DIMACS format is a standardized method of representing CNF formulas
        as strings. This implementation is based on the details given `here
        <https://people.sc.fsu.edu/~jburkardt/data/cnf/cnf.html>`_.
        """
        if fresh_variable_count is None:
            fresh_variable_count = self._num_vars
        header = f"p cnf {fresh_variable_count} {len(self)}\n\n"
        return header + str(self)

    def as_unigen_string(self,
                         fresh_variable_count: Optional[int] = None,
                         support_set_length: Optional[int] = None,
                         sampled_variables: Optional[List[Var]] = None) -> str:
        """Returns a string representing the CNF formula in the modified DIMACS
        format used by Unigen.

        See :meth:`CNF.as_dimacs_string` for details about the DIMACS format.

        Unigen adds an extra line at the top that describes the list of
        variables you would like to sample (instead of sampling among all
        variables). This line is given in the form::

            c ind v1 v2 ... vn 0

        where ``v1``, ``v2``, ..., ``vn`` represent variables in the formula.
        This line is placed just below the "problem" line (the line beginning
        with ``p``).
        """
        if support_set_length is not None and sampled_variables is not None:
            raise ValueError("cannot give both a support set length and sampled variables list to as_unigen_string!")
        elif support_set_length is not None:
            support_set = [Var(n) for n in range(1, support_set_length + 1)]
        elif sampled_variables is not None:
            support_set = sampled_variables
        else:
            support_set = []

        # This fun list comprehension divides the list of variables in the
        # support set into separate lists of no more than ten variables each,
        # due to restrictions in the file format.
        support_chunks = [[n for n in support_set[idx:idx + 10]] for idx in range(0, len(support_set), 10)]
        support_string = '\n'.join("c ind " + ' '.join(map(str, chunk)) + " 0"
                                   for chunk in support_chunks)

        # Now we modify the base DIMACS string by inserting the Unigen
        # modifications after the "problem" line. We know the problem line is
        # the first line, so we just do a simple substitution on the first
        # newline character in the string.
        dimacs_string = self.as_dimacs_string(fresh_variable_count)
        unigen_string = dimacs_string.replace('\n', '\n' + support_string, 1)

        # Done!
        return unigen_string

    def as_list_of_list_of_ints(self) -> List[List[int]]:
        """Converts the :class:`CNF` to a :class:`list` of
        :class:`lists <list>` of :class:`ints <int>`.
        """
        return [[int(var) for var in clause] for clause in self]

    def as_haskell_cnf(self) -> Tuple[int, List[List[int]]]:
        """Converts the :class:`CNF` to a :class:`tuple` whose first element is
        the number of fresh variables in the formula and whose second element
        is the :class:`CNF` represented as a :class:`list` of
        :class:`lists <list>` of :class:`ints <int>`.

        .. note::

            This is the way CNF formulas were encoded in the original Haskell
            code. It only exists for consistency checks across the two code
            bases.
        """
        return (self._num_vars, self.as_list_of_list_of_ints())

    ########################################
    ##
    ## Operator Overloads
    ##

    # CNF + ___
    def __add__(self, other: Union[CNF, Clause, Var]) -> CNF:
        """Logical OR. This alias exists due to the :class:`list`-like
        interface of :class:`CNFs <.CNF>`.
        """
        if isinstance(other, CNF):
            return CNF(*self, *other)
        if isinstance(other, Clause):
            return CNF(*self, other)
        if isinstance(other, Var):
            return CNF(*self, [other])
        return NotImplemented

    # CNF += ___
    def __iadd__(self, other: Union[CNF, Clause, Iterable[Clause], Var]) -> CNF:
        if isinstance(other, CNF):
            self._vals += other._vals
            return self
        if isinstance(other, Clause):
            self._vals += [other]
            return self
        if isinstance(other, (list, tuple)):
            self._vals += other
            return self
        if isinstance(other, Var):
            self._vals += [Clause(other)]
            return self
        return NotImplemented

    # CNF & ___
    def __and__(self, other: Union[Clause, Var]) -> CNF:
        """Logical AND."""
        if isinstance(other, Clause):
            return CNF(self._vals + [other])
        return CNF(self._vals + [Clause(other)])

    # ___ & CNF
    def __rand__(self, other: Union[Clause, Var]) -> CNF:
        if isinstance(other, Clause):
            return CNF([other] + self._vals)
        return CNF([Clause(other)] + self._vals)

    # CNF | ___
    def __or__(self, other: Var) -> CNF:
        """Logical OR."""
        return CNF([*self[:-1], self[-1] + other])

    # ___ | CNF
    def __ror__(self, other: Var) -> CNF:
        return CNF([other + self[0], *self[1:]])

    # CNF ** ___
    def __pow__(self, other: Var) -> CNF:
        """Distribution of a :class:`Var` across the :class:`Clauses <.Clause>`
        of a :class:`CNF`.
        """
        if isinstance(other, Var):
            return CNF([clause | other for clause in self])
        return NotImplemented

    # ___ ** CNF
    def __rpow__(self, other: Var) -> CNF:
        if isinstance(other, Var):
            return CNF([other | clause for clause in self])
        return NotImplemented

    ########################################
    ##
    ## Variable Manipulation Functions
    ##

    def get_fresh(self) -> Var:
        """Creates a new :class:`Var` for the formula."""
        # NOTE: I think this is a weird interface. A new variable is generated,
        #       permanently affecting this CNF formula... but there's no
        #       guarantee that the caller actually incorporates that variable
        #       into the formula.
        self._num_vars += 1
        return Var(self._num_vars)

    def get_n_fresh(self, n: int) -> List[Var]:
        """Creates the next *n* :class:`Vars <.Var>`, numbered sequentially."""
        return [self.get_fresh() for _ in range(n)]

    def append(self, other: Union[CNF, Clause, Iterable[Clause], Var]):
        """Appends a :class:`CNF` to this :class:`CNF`."""
        self += other

    def prepend(self, other: Union[CNF, Clause, Iterable[Clause], Var]):
        """Prepends a :class:`CNF` to this :class:`CNF`."""
        if isinstance(other, Var):
            self._vals.append(Clause(other))
        elif isinstance(other, Clause):
            self._vals.append(other)
        elif isinstance(other, CNF):
            self._vals.extend(other._vals)
        else:
            raise NotImplementedError()

    def set_to_zero(self, variable: Var):
        """Zeroes the specified :class:`Var` by appending its negation to the
        existing CNF formula.
        """
        self.prepend(~variable)

    def zero_out(self, in_list: Iterable[Var]):
        """Appends a CNF formula negating the existing CNF formula."""
        zeroed_cnf = CNF([[~var] for var in in_list])
        self.prepend(zeroed_cnf)

    def set_to_one(self, variable: Var):
        """Sets the specified variable to ``1`` by appending it to the existing
        CNF formula.
        """
        self.prepend(variable)

    ########################################
    ##
    ## CNF Assertions
    ##

    def assert_k_of_n(self, k: int, in_list: Sequence[Var]):
        # TODO DOC
        # TODO: Describe this function's purpose.
        in_binary =  int_to_binary(k)
        sum_bits = self.pop_count(in_list, len(in_binary)+1)
        # Add zero padding to the left.
        in_binary.reverse()
        left_padded: BinaryNumber = in_binary[:len(sum_bits)]
        left_padded += [-1 for _ in range(len(sum_bits) - len(left_padded))]
        left_padded.reverse()
        # Form the assertion.
        assertion = [Var(lp * sb.value) for (lp, sb) in zip(left_padded, sum_bits)]
        # Append the assertion to the formula.
        self.prepend(CNF([Clause(x) for x in assertion]))

    def assert_k_less_than_n(self, k: int, in_list: Sequence[Var]):
        # TODO DOC
        self._inequality_assertion(True, k, in_list)

    def assert_k_greater_than_n(self, k: int, in_list: Sequence[Var]):
        # TODO DOC
        self._inequality_assertion(False, k, in_list)

    def _inequality_assertion(self, assert_less_than: bool, k: int, in_list: Sequence[Var]):
        in_binary = int_to_binary(k)
        sum_bits = self.pop_count(in_list, len(in_binary)+1)
        k_vars = self.get_n_fresh(len(in_binary))
        assertion = [Var(kv.value * b) for (kv, b) in zip(k_vars, in_binary)]
        self.prepend(CNF([Clause(x) for x in assertion]))
        self._make_same_length(k_vars, sum_bits)
        if assert_less_than:
            kbs, nbs = sum_bits, k_vars
        else:
            kbs, nbs = k_vars, sum_bits
        neg_twos_comp_nbs = self._convert_to_negative_twos_complement(nbs)
        (_, ss) = self.ripple_carry(kbs, neg_twos_comp_nbs)
        self.set_to_one(ss[-1])

    def _make_same_length(self, xs: List[Var], ys: List[Var]):
        if len(xs) == len(ys):
            return
        elif len(xs) < len(ys):
            zero_padding = self.get_n_fresh(len(ys) - len(xs) + 1)
            self.zero_out(zero_padding)
            xs[:0] = zero_padding
            one_more_zero = self.get_n_fresh(1)
            self.zero_out(one_more_zero)
            ys[:0] = one_more_zero
        else:
            self._make_same_length(ys, xs)

    def _convert_to_negative_twos_complement(self, bits: List[Var]) -> List[Var]:
        # Flip the bits, i.e., assert flipped_bits[i] ⇔ bits[i].
        flipped_bits = self.get_n_fresh(len(bits))
        flipped_cnf = CNF()
        for lhs, rhs in zip(flipped_bits, (~b for b in bits)):
            double_implied = CNF.xnor_vars(lhs, rhs)
            flipped_cnf += double_implied
        self.prepend(flipped_cnf)
        # Make a zero-padded one (for the addition) of the correct dimension.
        one_vars = self.get_n_fresh(len(bits))
        # Set all the top bits to 0 and the bottom bit to 1.
        self.zero_out(one_vars[:-1])
        self.set_to_one(one_vars[-1])
        # Add the lists.
        (_, ss) = self.ripple_carry(flipped_bits, one_vars)
        ss.reverse()
        return ss

    ########################################
    ##
    ## Pop Count
    ##

    def pop_count(self, in_list: Sequence[Var], saturate_at: int) -> List[Var]:
        """Returns the list of :class:`Vars <.Var>` that represents the bits of
        the given list variable in binary.
        """
        if not in_list:
            raise ValueError("cannot take pop count of empty list")
        # Our pop count algorithm assumes the input is a binary number with a
        # power-of-two number of digits. So first, we find out what the next
        # power of two is for our number (which may be just the length of that
        # number).
        nearest_largest_power = math.ceil(math.log(len(in_list), 2))
        # Then we generate a sequence of fresh variables (new digits) equal to
        # the difference and set them all to 0.
        aux_list = self.get_n_fresh((2 ** nearest_largest_power) - len(in_list))
        self.zero_out(aux_list)
        # Now we can start computing the actual pop count.
        return self._pop_count_layer([[x] for x in chain(in_list, aux_list)], saturate_at)

    def _pop_count_layer(self, bit_list: List[List[Var]], saturate_at: int) -> List[Var]:
        if len(bit_list) == 1:
            return bit_list[0]
        midpoint = len(bit_list) // 2
        left_half = bit_list[:midpoint]
        right_half = bit_list[midpoint:]
        var_list: List[List[Var]] = []
        # This zip assumes the two lists are of equal length. This is a safe
        # assumption since we've guaranteed the input to have a length that's a
        # power of two greater than one.
        for (l, r) in zip(left_half, right_half):
            if saturate_at == 0:
                (max_c, ss) = self.ripple_carry(l, r)
                var_list.append([max_c] + list(reversed(ss)))
            else:
                var_list.append(self.ripple_saturate(l, r, saturate_at))
        # Reverse the list because of append ordering.
        var_list.reverse()
        return self._pop_count_layer(var_list, saturate_at)

    ########################################
    ##
    ## Adders
    ##

    def half_adder(self, a: Var, b: Var) -> Tuple[Var, Var]:
        # TODO DOC
        c = self.get_fresh()
        s = self.get_fresh()

        c_val = CNF.and_vars(a, b)
        c_neg_val = CNF.or_vars(~a, ~b)
        c_implies_c_val = CNF.distribute(~c, c_val)
        c_val_implies_c = CNF.distribute(c, c_neg_val)
        computed_c = c_implies_c_val + c_val_implies_c
        self.prepend(computed_c)

        s_val = CNF.xor_vars(a, b)
        s_neg_val = CNF.xnor_vars(a, b)
        s_implies_s_val = CNF.distribute(~s, s_val)
        s_val_implies_s = CNF.distribute(s, s_neg_val)
        computed_s = s_implies_s_val + s_val_implies_s
        self.prepend(computed_s)

        return (c, s)

    def full_adder(self, a: Var, b: Var, cin: Optional[Var]) -> Tuple[Var, Var]:
        # TODO DOC

        if not cin:
            return self.half_adder(a, b)
        
        cout = self.get_fresh()
        s = self.get_fresh()

        c_val     = (a | b) & (a | cin) & (b | cin)
        c_neg_val = (~a | ~b) & (~a | ~cin) & (~b | ~cin)
        c_implies_c_val = CNF.distribute(~cout, c_val)
        c_val_implies_c = CNF.distribute(cout, c_neg_val)
        computed_c = c_implies_c_val + c_val_implies_c
        self.prepend(computed_c)

        s_val     = (~a | ~b | cin) & (~a | b | ~cin) & (a | ~b | ~cin) & (a | b | cin)
        s_neg_val = (~a | ~b | ~cin) & (~a | b | cin) & (a | ~b | cin) & (a | b | ~cin)
        s_implies_s_val = CNF.distribute(~s, s_val)
        s_val_implies_s = CNF.distribute(s, s_neg_val)
        computed_s = s_implies_s_val + s_val_implies_s
        self.prepend(computed_s)

        return (cout, s)

    def saturate_adder(self, a: Var, b: Var, cin: Optional[Var]) -> Var:
        # TODO DOC
        s = self.get_fresh()

        if (cin):
            s_val     = CNF(a | b | cin)
            s_neg_val = (~a & ~b & ~cin)
        else:
            s_val     = CNF(a | b)
            s_neg_val = (~a & ~b)
        s_implies_s_val = CNF.distribute(~s, s_val)
        s_val_implies_s = CNF.distribute(s, s_neg_val)
        computed_s = s_implies_s_val + s_val_implies_s
        self.prepend(computed_s)

        return s

    def ripple_carry(self, xs: List[Var], ys: List[Var]) -> Tuple[Var, List[Var]]:
        # TODO DOC
        cin = None

        s_accum: List[Var] = []

        for x, y in zip(reversed(xs), reversed(ys)):
            (c, s) = self.full_adder(x, y, cin)
            s_accum.append(s)
            cin = c

        return (cast(Var, cin), s_accum)

    def ripple_saturate(self, xs: List[Var], ys: List[Var], saturate_at: int) -> List[Var]:
        # Assuming xs and ys have no more than `saturate_at` variables,
        # generate a saturating sum with no more than `saturate_at` variables
        cin = None

        s_accum: List[Var] = []

        for i, (x, y) in enumerate(zip(reversed(xs), reversed(ys))):
            if i+1 == saturate_at:
                s = self.saturate_adder(x, y, cin)
                s_accum.append(s)
            else:
                (c, s) = self.full_adder(x, y, cin)
                s_accum.append(s)
                cin = c

        if len(xs) < saturate_at:
            s_accum.append(cast(Var, cin))

        s_accum.reverse()
        return s_accum



"""This module provides the base constraint class."""


from abc import ABC, abstractmethod
from typing import List

from sweetpea._internal.primitive import Factor


class Constraint(ABC):
    """Generic interface for constraints."""

    @abstractmethod
    def validate(self, block) -> None:
        """Constraints can't be completely validated in isolation. This
        function will be called on all constraints with the block during the
        block initialization to ensure all constraints are valid.
        """
        pass

    @abstractmethod
    def apply(self, block, backend_request) -> None:
        pass

    def is_complex_for_combinatoric(self) -> bool:
        return True

    def desugar(self, replacements: dict) -> List:
        """Some constraints accept shorthand representations. (Like accepting a
        whole factor, rather than individual factor and level name pairs.)

        :func:`.Constraint.desugar` is responsible for returning the desugared
        representation of a constraint. When a block is constructed, it will
        desugar all constraints, replacing them with their desuagared versions
        before proceding.
        """
        return [self]

    def uses_factor(self, f: Factor) -> bool:
        """Reports whether the given factor is relevant to the constraint, and
        can return False when a factor is known to be relevant to the
        crossing or some other constraint.
        """
        return False

    @abstractmethod
    def potential_sample_conforms(self, sample: dict, block) -> bool:
        """For rejection sampling, checks whether a given potential sample
        matches a constraint, as long as crossing, exclusion for non-complex factors, and
        minimum-trial contraints are already satisfied.
        """
        pass

    def set_within_block(self) -> None:
        pass



from sweetpea._internal.logic import If, Iff, And, Or, Not, to_cnf_naive, to_cnf_switching, to_cnf_tseitin, cnf_to_json


def test_to_cnf_naive():
    assert to_cnf_naive(1, 2) == (And([1]), 2)
    assert to_cnf_naive(And([1]), 2) == (And([1]), 2)
    assert to_cnf_naive(And([1, 2]), 3) == (And([1, 2]), 3)

    assert to_cnf_naive(Or([1, 2]), 3) == (And([Or([1, 2])]), 3)
    assert to_cnf_naive(Or([1, And([2, 3])]), 4) == (And([Or([1, 2]), Or([1, 3])]), 4)

    formula = And([
        Iff(1, And([2, 3])),
        Iff(4, And([5, 6]))
    ])
    expected_cnf = And([
        Or([1, Not(2), Not(3)]),
        Or([Not(1), 2]),
        Or([Not(1), 3]),
        Or([4, Not(5), Not(6)]),
        Or([Not(4), 5]),
        Or([Not(4), 6])
    ])
    assert to_cnf_switching(formula, 7) == (expected_cnf, 7)


def test_to_cnf_switching():
    formula = Or([
        And([3, 4]),
        And([Not(2), Or([1, 5])])
    ])
    expected_cnf = And([
        Or([3, Not(6)]),
        Or([4, Not(6)]),
        Or([1, 5, 6]),
        Or([Not(2), 6])
    ])
    assert to_cnf_switching(formula, 6) == (expected_cnf, 7)

    formula = And([
        Iff(1, And([2, 3])),
        Iff(4, And([5, 6]))
    ])
    expected_cnf = And([
        Or([1, Not(2), Not(3)]),
        Or([Not(1), 2]),
        Or([Not(1), 3]),
        Or([4, Not(5), Not(6)]),
        Or([Not(4), 5]),
        Or([Not(4), 6])
    ])
    assert to_cnf_switching(formula, 7) == (expected_cnf, 7)


def test_to_cnf_tseitin():
    assert to_cnf_tseitin(Or([1, And([2, 3])]), 4) == (And([
        # 4 <=> (2 ^ 3)
        Or([Not(2), Not(3), 4]),
        Or([2, Not(4)]),
        Or([3, Not(4)]),

        # 5 <=> (1 v 4)
        Or([1, 4, Not(5)]),
        Or([Not(1), 5]),
        Or([Not(4), 5]),

        # Final clause
        5
    ]), 6)


def test_cnf_to_json():
    assert cnf_to_json([And([1])]) == [[1]]

    assert cnf_to_json([And([Or([1])])]) == [[1]]
    assert cnf_to_json([And([Or([Not(5)])])]) == [[-5]]
    assert cnf_to_json([And([Or([1, 2, Not(4)])])]) == [[1, 2, -4]]

    assert cnf_to_json([And([Or([1, 4]), Or([5, -4, 2]), Or([-1, -5])])]) == [
        [1, 4],
        [5, -4, 2],
        [-1, -5]]

    assert cnf_to_json([And([
        Or([1, 3, Not(2)]),
        Or([1, 3, 5]),
        Or([1, 4, Not(2)]),
        Or([1, 4, 5]),
        Or([2, 3, 1, 5]),
        Or([2, 4, 1, 5])])]) == [
            [1, 3, -2],
            [1, 3, 5],
            [1, 4, -2],
            [1, 4, 5],
            [2, 3, 1, 5],
            [2, 4, 1, 5]]


def test_eliminate_iff():
    from sweetpea._internal.logic import __eliminate_iff

    # P -> Q  ==> (~P v Q)
    assert __eliminate_iff(If(1, 2)) == Or([Not(1), 2])

    # P <-> Q ==> (P v ~Q) ^ (~P v Q)
    assert __eliminate_iff(Iff(1, 2)) == And([Or([1, Not(2)]), Or([Not(1), 2])])

    assert __eliminate_iff(Iff(1, And([2, 3]))) == And([
        Or([1, Not(And([2, 3]))]),
        Or([Not(1), And([2, 3])])
    ])


def test_apply_demorgan():
    from sweetpea._internal.logic import __apply_demorgan

    # P ==> P, ~P ==> ~P
    assert __apply_demorgan(4) == 4
    assert __apply_demorgan(Not(4)) == Not(4)

    # ~~P ==> P
    assert __apply_demorgan(Not(Not(4))) == 4

    # ~(P v Q) ==> ~P ^ ~Q
    assert __apply_demorgan(Not(Or([1, 2]))) == And([Not(1), Not(2)])

    # ~(P ^ Q) ==> ~P v ~Q
    assert __apply_demorgan(Not(And([1, 2]))) == Or([Not(1), Not(2)])

    assert __apply_demorgan(Or([1, Not(And([2, 3]))])) == Or([1, Not(2), Not(3)])


def test_distribute_ors_naive():
    from sweetpea._internal.logic import __distribute_ors_naive

    # When f is int or Not, return it. (Not can only contain int, as we've
    # already applied DeMorgan's laws)
    assert __distribute_ors_naive(1) == 1
    assert __distribute_ors_naive(Not(1)) == Not(1)

    # When f in an Or, distribute the Or over the contained clauses.
    assert __distribute_ors_naive(Or([1, 2])) == And([Or([1, 2])])
    assert __distribute_ors_naive(Or([1, And([2, 3])])) == And([Or([1, 2]), Or([1, 3])])
    assert __distribute_ors_naive(Or([And([1, 2]), And([3, 4])])) == And([
        Or([1, 3]), Or([1, 4]), Or([2, 3]), Or([2, 4])
    ])

    # When f is an And, disitribute Ors over the contained clauses.
    assert __distribute_ors_naive(And([1, Not(2)])) == And([1, Not(2)])
    assert __distribute_ors_naive(And([1, Or([2, And([3, 4])])])) == And([
        Or([2, 3]), Or([2, 4]), 1
    ])


def test_distribute_ors_switching():
    from sweetpea._internal.logic import __distribute_ors_switching

    # When lhs or rhs is a single variable, just distribute it.
    assert __distribute_ors_switching(Or([1, And([2, 3])]), 4) == (
        And([Or([1, 2]), Or([1, 3])]),
        4
    )

    assert __distribute_ors_switching(Or([And([1, 2]), 3]), 4) == (
        And([Or([1, 3]), Or([2, 3])]),
        4
    )

    # Should distribute over multiple individual variables
    assert __distribute_ors_switching(Or([
        1, 2, And([3, 4])
    ]), 5) == (
        And([
            Or([1, 2, 3]),
            Or([1, 2, 4])
        ]),
        5
    )

    # When both the lhs and rhs are more than just a single variable,
    # then introduce a switching variable to limit the formula growth.
    assert __distribute_ors_switching(Or([And([1, 2]), And([3, 4])]), 5) == (
        And([
            Or([1, Not(5)]),
            Or([2, Not(5)]),
            Or([3, 5]),
            Or([4, 5])
        ]),
        6
    )


def test_tseitin_rep_variables():
    from sweetpea._internal.logic import __tseitin_rep, _Cache

    clauses = []
    cache = _Cache(2)
    assert __tseitin_rep(1, clauses, cache) == 1
    assert clauses == []
    assert cache.get_next_variable() == 2


def test_tseitin_rep_not():
    from sweetpea._internal.logic import __tseitin_rep, _Cache

    # Should replace Not(var) with another variable
    clauses = []
    cache = _Cache(2)
    assert __tseitin_rep(Not(1), clauses, cache) == 2

    # Make sure that Not(1) was cached.
    assert cache.get(str(Not(1))) == 2

    # Make sure that the correct implication clauses were added.
    assert Or([    1,      2 ]) in clauses
    assert Or([Not(1), Not(2)]) in clauses

    # No clauses should be added if the value was already cached.
    clauses = []
    cache = _Cache(2)
    assert cache.get(str(Not(1))) == 2 # Prewarm the cache.
    assert __tseitin_rep(Not(1), clauses, cache) == 2
    assert clauses == []


def test_tseitin_rep_if():
    from sweetpea._internal.logic import __tseitin_rep, _Cache

    clauses = []
    cache = _Cache(3)

    # Make sure return is correct and value was cached.
    assert __tseitin_rep(If(1, 2), clauses, cache) == 3
    assert cache.get(str(If(1, 2))) == 3

    # Make sure equivalence clauses were added.
    assert Or([Not(1),     2,  Not(3)]) in clauses
    assert Or([    1,      3]) in clauses
    assert Or([Not(2),     3]) in clauses

    # Don't duplicate clauses when the cache was already populated.
    clauses = []
    cache = _Cache(3)

    # Prewarm the cache.
    assert cache.get(str(If(1, 2))) == 3
    assert __tseitin_rep(If(1, 2), clauses, cache) == 3

    # Make sure no clauses were added.
    assert clauses == []


def test_tseitin_rep_iff():
    from sweetpea._internal.logic import __tseitin_rep, _Cache

    clauses = []
    cache = _Cache(3)

    # Make sure return is correct and value was cached.
    assert __tseitin_rep(Iff(1, 2), clauses, cache) == 3
    assert cache.get(str(Iff(1, 2))) == 3

    # Make sure equivalence clauses were added.
    assert Or([    1,      2,      3 ]) in clauses
    assert Or([Not(1), Not(2),     3 ]) in clauses
    assert Or([    1,  Not(2), Not(3)]) in clauses
    assert Or([Not(1),     2,  Not(3)]) in clauses

    # Don't duplicate clauses when the cache was already populated.
    clauses = []
    cache = _Cache(3)

    # Prewarm the cache.
    assert cache.get(str(Iff(1, 2))) == 3
    assert __tseitin_rep(Iff(1, 2), clauses, cache) == 3

    # Make sure no clauses were added.
    assert clauses == []


def test_tseitin_rep_and():
    from sweetpea._internal.logic import __tseitin_rep, _Cache

    clauses = []
    cache = _Cache(4)

    # Make sure return is correct, and value was cached.
    assert __tseitin_rep(And([1, 2, 3]), clauses, cache) == 4
    assert cache.get(str(And([1, 2, 3]))) == 4

    # Make sure equivalence clauses were added.
    assert Or([1, Not(4)]) in clauses
    assert Or([2, Not(4)]) in clauses
    assert Or([3, Not(4)]) in clauses
    assert Or([Not(1), Not(2), Not(3), 4]) in clauses

    # Don't duplicate clauses when the cache was already populated
    clauses = []
    cache = _Cache(4)

    # Prewarm the cache
    assert cache.get(str(And([1, 2, 3]))) == 4
    assert __tseitin_rep(And([1, 2, 3]), clauses, cache) == 4

    # Make sure no clauses were added
    assert clauses == []


def test_tseitin_rep_or():
    from sweetpea._internal.logic import __tseitin_rep, _Cache

    clauses = []
    cache = _Cache(4)

    # Make sure return is correct, and value was cached.
    assert __tseitin_rep(Or([1, 2, 3]), clauses, cache) == 4
    assert cache.get(str(Or([1, 2, 3]))) == 4

    # Make sure equivalence clauses were added.
    assert Or([Not(1), 4]) in clauses
    assert Or([Not(2), 4]) in clauses
    assert Or([Not(3), 4]) in clauses
    assert Or([1, 2, 3, Not(4)]) in clauses

    # Don't duplicate clauses when the cache was already populated
    clauses = []
    cache = _Cache(4)

    # Prewarm the cache
    assert cache.get(str(Or([1, 2, 3]))) == 4
    assert __tseitin_rep(Or([1, 2, 3]), clauses, cache) == 4

    # Make sure no clauses were added
    assert clauses == []


def test_tseitin_cache():
    from sweetpea._internal.logic import _Cache

    c = _Cache(5)

    assert c.get('x') == 5
    assert c.get('x') == 5
    assert c.get('y') == 6
    assert c.get('z') == 7
    assert c.get('x') == 5

    assert c.get_next_variable() == 8



"""This module provides functionality for calling the third-party Unigen tool.

`Unigen <https://github.com/meelgroup/unigen>`_ is a state-of-the-art,
almost-uniform SAT sampler that uses `CryptoMiniSAT
<https://github.com/msoos/cryptominisat>`_ to solve SAT problems. SweetPea uses
Unigen for a few processes.
"""


from pathlib import Path
from shlex import split as shell_split
from subprocess import CompletedProcess, run
from numpy import random
from tempfile import NamedTemporaryFile

from typing import Tuple

from .docker_utility import DEFAULT_DOCKER_MODE_ON, docker_run
from .executables import DEFAULT_DOWNLOAD_IF_MISSING, UNIGEN_EXE, CMSGEN_EXE, ensure_executable_available
from .tool_error import ToolError
from ..utility import temporary_cnf_file


__all__ = ['DEFAULT_DOCKER_MODE_ON', 'UnigenError', 'call_unigen']


class UnigenError(ToolError):
    """An error raised when Unigen fails."""
    pass


def call_unigen_docker(input_file: Path, sample_count: int) -> Tuple[CompletedProcess, str]:
    """Calls Unigen in a Docker container, reading a given file as the input
    problem.
    """
    unigen_container = 'msoos/unigen'
    input_bytes = input_file.read_bytes()
    # args = shell_split("--rm -i -a stdin -a stdout --samples="+str(sample_count))
    args = shell_split("--rm -i -a stdin -a stdout")
    result = docker_run(unigen_container, args, input_bytes)
    return (result, "")


def call_unigen_cli(input_file: Path,
                    download_if_missing: bool,
                    sample_count: int,
                    use_cmsgen: bool) -> Tuple[CompletedProcess, str]:
    """Calls Unigen from the command line, reading a given file as the input
    problem.

    If ``download_if_missing`` is ``True``, SweetPea will automatically
    download the Unigen executable (and other executables SweetPea depends on)
    to a local directory from the `sweetpea-org/unigen-exe repository
    <https://github.com/sweetpea-org/unigen-exe>`_.
    """
    unigen_exe = CMSGEN_EXE if use_cmsgen else UNIGEN_EXE
    ensure_executable_available(unigen_exe, download_if_missing)
    seed = random.randint(999999999)
    command = [str(unigen_exe), str(input_file), "--samples="+str(sample_count), "--seed="+str(seed)]
    if use_cmsgen:
        with temporary_cnf_file(suffix = ".out") as output_file:
            command.append("--samplefile="+output_file.name)
            result = run(command, capture_output=True)  # noqa
            with open(output_file, 'r') as output:
                samples = output.read()
                return (result, samples)
    else:
        # NOTE: flake8 doesn't seem to handle the calls to `run` correctly, but
        #       mypy reports everything is fine here so we `noqa` to prevent flake8
        #       complaining about what it doesn't understand.
        result = run(command, capture_output=True)  # noqa
        return (result, "")


def call_unigen(sample_count: int,
                input_file: Path,
                docker_mode: bool = DEFAULT_DOCKER_MODE_ON,
                download_if_missing: bool = DEFAULT_DOWNLOAD_IF_MISSING,
                use_cmsgen: bool = False
                ) -> str:
    """Calls Unigen with the given file as input.

    If ``docker_mode`` is ``True``, this will use a Docker container to run
    Unigen. If it's ``False``, a command-line executable will be used.

    If ``docker_mode`` is ``False`` and no local Unigen executable can be
    found, and if ``download_if_missing`` is ``True``, the needed executable
    will be automatically downloaded if it's missing.
    """
    if docker_mode:
        (result, samples) = call_unigen_docker(input_file, sample_count)
    else:
        (result, samples) = call_unigen_cli(input_file, download_if_missing, sample_count, use_cmsgen)
    if result.returncode == (10 if use_cmsgen else 0):
        # Success!
        # (Comments in the earlier Haskell version of SweetPea's core indicate
        # that Unigen used to use 0 as an error indicator.)
        return result.stdout.decode()+samples
    else:
        # Failure.
        stdout = result.stdout.decode()
        stderr = result.stderr.decode()
        if "The input formula is unsatisfiable" in stdout:
            return ""
        raise UnigenError(result.returncode, stdout, stderr)



from sweetpea._internal.sampling_strategy.base import Gen, SamplingResult
from sweetpea._internal.block import Block
from sweetpea._internal.sampling_strategy.iterate_sat import IterateSATGen
from sweetpea._internal.sampling_strategy.random import RandomGen

"""
This represents a strategy where we "sample" just by using some
solver repeatedly to produce unique (but not necessarily uniform) samples.
A specific strategy is selected automatically.
"""
class IterateGen(Gen):
    
    @staticmethod
    def class_name():
        return 'NonUniformGen'

    @staticmethod
    def sample(block: Block, sample_count: int) -> SamplingResult:
        if block.complex_factors_or_constraints:
            return IterateSATGen.sample(block, sample_count)
        else:
            return RandomGen.sample(block, sample_count)



from sweetpea._internal.sampling_strategy.base import Gen, SamplingResult
from sweetpea._internal.block import Block
from sweetpea._internal.core import CNF
from sweetpea._internal.core.generate.sample_ilp import sample_ilp_iterate

class IterateILPGen(Gen):

    @staticmethod
    def class_name():
        return 'IterateILPGen'

    @staticmethod
    def sample(block: Block, sample_count: int) -> SamplingResult:
        backend_request = block.build_backend_request()
        if block.show_errors():
            return SamplingResult([], {})

        solutions = sample_ilp_iterate(sample_count,
                                       CNF(backend_request.get_cnfs_as_json()),
                                       block.variables_per_sample(),
                                       backend_request.get_requests_as_generation_requests())

        result = list(map(lambda s: Gen.decode(block, s.assignment), solutions))
        return SamplingResult(result, {})



import operator as op
import pytest

from sweetpea import CrossBlock
from sweetpea._internal.primitive import Factor, DerivedLevel, WithinTrial
from sweetpea._internal.design_partition import DesignPartitions
from sweetpea._internal.constraint import Reify


color      = Factor("color",      ["red", "blue"])
text       = Factor("text",       ["red", "blue"])
congruency = Factor("congruency", [
    DerivedLevel("congruent",   WithinTrial(op.eq, [color, text])),
    DerivedLevel("incongruent", WithinTrial(op.ne, [color, text]))
])
color_red  = Factor("color red", [
    DerivedLevel("yes", WithinTrial(lambda c: c == "red", [color])),
    DerivedLevel("no",  WithinTrial(lambda c: c != "red", [color]))
])

design   = [color, text, congruency, color_red]
crossing = [color, congruency]
block    = CrossBlock(design, crossing, list(map(Reify, design)))

def test_get_crossed_factors():
    partitions = DesignPartitions(block)
    assert partitions.get_crossed_noncomplex_factors() == crossing


def test_get_crossed_factors_derived():
    partitions = DesignPartitions(block)
    assert partitions.get_crossed_noncomplex_derived_factors() == [congruency]


def test_get_uncrossed_basic_factors():
    partitions = DesignPartitions(block)
    assert partitions.get_uncrossed_basic_factors() == [text]


def test_get_uncrossed_basic_source_factors():
    partitions = DesignPartitions(block)
    assert partitions.get_uncrossed_basic_source_factors() == [text]


def test_get_uncrossed_basic_independent_factors():
    partitions = DesignPartitions(block)
    assert partitions.get_uncrossed_basic_independent_factors() == []


def test_get_uncrossed_derived_factors():
    partitions = DesignPartitions(block)
    assert partitions.get_uncrossed_derived_and_complex_derived_factors() == [color_red]






import operator as op
import pytest

from itertools import permutations

from sweetpea import CrossBlock
from sweetpea._internal.block import Block
from sweetpea._internal.primitive import Factor, DerivedLevel, WithinTrial, Transition, Window, SimpleLevel
from sweetpea._internal.constraint import Constraint, Consistency, Cross, Derivation, AtMostKInARow, ExactlyKInARow, AtLeastKInARow, Exclude, Pin, Reify
from sweetpea._internal.backend import LowLevelRequest, BackendRequest
from sweetpea._internal.logic import And, Or, If, Iff, Not, to_cnf_tseitin

color = Factor("color", ["red", "blue"])
text  = Factor("text",  ["red", "blue"])

con_level  = DerivedLevel("con", WithinTrial(op.eq, [color, text]))
inc_level  = DerivedLevel("inc", WithinTrial(op.ne, [color, text]))
con_factor = Factor("congruent?", [con_level, inc_level])

color_repeats_factor = Factor("color repeats?", [
    DerivedLevel("yes", Transition(lambda colors: colors[0] == colors[-1], [color])),
    DerivedLevel("no",  Transition(lambda colors: colors[0] != colors[-1], [color]))
])

text_repeats_factor = Factor("text repeats?", [
    DerivedLevel("yes", Transition(lambda texts: texts[0] == texts[-1], [text])),
    DerivedLevel("no",  Transition(lambda texts: texts[0] != texts[-1], [text]))
])

congruent_bookend = Factor("congruent bookend?", [
    DerivedLevel("yes", Window(lambda color, text: color == text, [color, text], 1, 3)),
    DerivedLevel("no",  Window(lambda color, text: color != text, [color, text], 1, 3))
])

design = [color, text, con_factor]
crossing = [color, text]
block = CrossBlock(design, crossing, [Reify(con_factor)])

def test_consistency():
    # From standard example
    # [ LowLevelRequest("EQ", 1, [1, 2]), LowLevelRequest("EQ", 1, [3, 4]), ...]
    backend_request = BackendRequest(0)

    Consistency.apply(block, backend_request)
    assert backend_request.ll_requests == \
        list(map(lambda x: LowLevelRequest("EQ", 1, [x, x+1]), range(1, 24, 2)))

    # Different case
    backend_request = BackendRequest(0)
    f = Factor("a", ["b", "c", "d", "e"])
    f_block = CrossBlock([f], [f], [])

    Consistency.apply(f_block, backend_request)
    assert backend_request.ll_requests == \
        list(map(lambda x: LowLevelRequest("EQ", 1, [x, x+1, x+2, x+3]), range(1, 16, 4)))

    # Varied level lengths
    backend_request = BackendRequest(0)
    f1 = Factor("a", ["b", "c", "d"])
    f2 = Factor("e", ["f"])
    f_block = CrossBlock([f1, f2], [f1, f2], [])

    Consistency.apply(f_block, backend_request)
    assert backend_request.ll_requests == [
        LowLevelRequest("EQ", 1, [1, 2, 3]), LowLevelRequest("EQ", 1, [4]),
        LowLevelRequest("EQ", 1, [5, 6, 7]), LowLevelRequest("EQ", 1, [8]),
        LowLevelRequest("EQ", 1, [9, 10, 11]), LowLevelRequest("EQ", 1, [12])]


@pytest.mark.parametrize('design', permutations([color, text, color_repeats_factor]))
def test_consistency_with_transition(design):
    block = CrossBlock(design, [color, text], [Reify(color_repeats_factor)])

    backend_request = BackendRequest(0)
    Consistency.apply(block, backend_request)

    # Because the color_repeats_factor doesn't apply to the first trial, (there isn't a previous trial
    # to compare to) the variables only go up to 22.
    assert backend_request.ll_requests == \
        list(map(lambda x: LowLevelRequest("EQ", 1, [x, x+1]), range(1, 22, 2)))


@pytest.mark.parametrize('design', permutations([color, text, color_repeats_factor, text_repeats_factor]))
def test_consistency_with_multiple_transitions(design):
    block = CrossBlock(design, [color, text], [Reify(color_repeats_factor),
                                               Reify(text_repeats_factor)])

    backend_request = BackendRequest(0)
    Consistency.apply(block, backend_request)

    assert backend_request.ll_requests == \
        list(map(lambda x: LowLevelRequest("EQ", 1, [x, x+1]), range(1, 28, 2)))


def test_consistency_with_transition_first_and_uneven_level_lengths():
    color3 = Factor("color3", ["red", "blue", "green"])

    yes_fn = lambda colors: colors[0] == colors[-1] == colors[-2]
    no_fn = lambda colors: not yes_fn(colors)
    color3_repeats_factor = Factor("color3 repeats?", [
        DerivedLevel("yes", Window(yes_fn, [color3], 3, 1)),
        DerivedLevel("no",  Window(no_fn, [color3], 3, 1))
    ])

    block = CrossBlock([color3_repeats_factor, color3, text], [color3, text],
                       [Reify(color3_repeats_factor)])

    backend_request = BackendRequest(0)
    Consistency.apply(block, backend_request)

    assert backend_request.ll_requests == [
        LowLevelRequest("EQ", 1, [1,  2,  3 ]), LowLevelRequest("EQ", 1, [4,  5 ]),
        LowLevelRequest("EQ", 1, [6,  7,  8 ]), LowLevelRequest("EQ", 1, [9,  10]),
        LowLevelRequest("EQ", 1, [11, 12, 13]), LowLevelRequest("EQ", 1, [14, 15]),
        LowLevelRequest("EQ", 1, [16, 17, 18]), LowLevelRequest("EQ", 1, [19, 20]),
        LowLevelRequest("EQ", 1, [21, 22, 23]), LowLevelRequest("EQ", 1, [24, 25]),
        LowLevelRequest("EQ", 1, [26, 27, 28]), LowLevelRequest("EQ", 1, [29, 30]),

        LowLevelRequest("EQ", 1, [31, 32]),
        LowLevelRequest("EQ", 1, [33, 34]),
        LowLevelRequest("EQ", 1, [35, 36]),
        LowLevelRequest("EQ", 1, [37, 38])
    ]


def test_consistency_with_general_window():
    design = [color, text, congruent_bookend]
    crossing = [color, text]
    block = CrossBlock(design, crossing, [Reify(congruent_bookend)])

    backend_request = BackendRequest(0)
    Consistency.apply(block, backend_request)

    assert backend_request.ll_requests == [
        LowLevelRequest("EQ", 1, [1,  2 ]), LowLevelRequest("EQ", 1, [3,  4 ]),
        LowLevelRequest("EQ", 1, [5,  6 ]), LowLevelRequest("EQ", 1, [7,  8 ]),
        LowLevelRequest("EQ", 1, [9,  10]), LowLevelRequest("EQ", 1, [11, 12]),
        LowLevelRequest("EQ", 1, [13, 14]), LowLevelRequest("EQ", 1, [15, 16]),

        LowLevelRequest("EQ", 1, [17, 18]),
        LowLevelRequest("EQ", 1, [19, 20])
    ]


def test_fully_cross_simple():
    block = CrossBlock([color, text],
                       [color, text],
                       [])

    (expected_cnf, _) = to_cnf_tseitin(And([
        Iff(17, And([1,  3 ])), Iff(18, And([1,  4 ])), Iff(19, And([2,  3 ])), Iff(20, And([2,  4 ])),
        Iff(21, And([5,  7 ])), Iff(22, And([5,  8 ])), Iff(23, And([6,  7 ])), Iff(24, And([6,  8 ])),
        Iff(25, And([9,  11])), Iff(26, And([9,  12])), Iff(27, And([10, 11])), Iff(28, And([10, 12])),
        Iff(29, And([13, 15])), Iff(30, And([13, 16])), Iff(31, And([14, 15])), Iff(32, And([14, 16]))
    ]), 33)

    backend_request = BackendRequest(17)
    Cross.apply(block, backend_request)

    assert backend_request.fresh == 66
    assert backend_request.cnfs == [expected_cnf]
    assert backend_request.ll_requests == [
        LowLevelRequest("EQ", 1, [17, 21, 25, 29]),
        LowLevelRequest("EQ", 1, [18, 22, 26, 30]),
        LowLevelRequest("EQ", 1, [19, 23, 27, 31]),
        LowLevelRequest("EQ", 1, [20, 24, 28, 32])
    ]


def test_fully_cross_with_constraint():
    (expected_cnf, _) = to_cnf_tseitin(And([
        Iff(25, And([1,  3 ])), Iff(26, And([1,  4 ])), Iff(27, And([2,  3 ])), Iff(28, And([2,  4 ])),
        Iff(29, And([7,  9 ])), Iff(30, And([7,  10])), Iff(31, And([8,  9 ])), Iff(32, And([8,  10])),
        Iff(33, And([13, 15])), Iff(34, And([13, 16])), Iff(35, And([14, 15])), Iff(36, And([14, 16])),
        Iff(37, And([19, 21])), Iff(38, And([19, 22])), Iff(39, And([20, 21])), Iff(40, And([20, 22]))
    ]), 41)

    backend_request = BackendRequest(25)
    Cross.apply(block, backend_request)

    assert backend_request.fresh == 74
    assert backend_request.cnfs == [expected_cnf]
    assert backend_request.ll_requests == [
        LowLevelRequest("EQ", 1, [25, 29, 33, 37]),
        LowLevelRequest("EQ", 1, [26, 30, 34, 38]),
        LowLevelRequest("EQ", 1, [27, 31, 35, 39]),
        LowLevelRequest("EQ", 1, [28, 32, 36, 40])
    ]


@pytest.mark.parametrize('design',
    [[color, text, color_repeats_factor],
     [color_repeats_factor, color, text]])
def test_fully_cross_with_transition_in_design(design):
    block = CrossBlock(design,
                       [color, text],
                       list(map(Reify, design)))

    backend_request = BackendRequest(23)
    Cross.apply(block, backend_request)

    (expected_cnf, _) = to_cnf_tseitin(And([
        Iff(23, And([1,  3 ])), Iff(24, And([1,  4 ])), Iff(25, And([2,  3 ])), Iff(26, And([2,  4 ])),
        Iff(27, And([5,  7 ])), Iff(28, And([5,  8 ])), Iff(29, And([6,  7 ])), Iff(30, And([6,  8 ])),
        Iff(31, And([9,  11])), Iff(32, And([9,  12])), Iff(33, And([10, 11])), Iff(34, And([10, 12])),
        Iff(35, And([13, 15])), Iff(36, And([13, 16])), Iff(37, And([14, 15])), Iff(38, And([14, 16]))
    ]), 39)

    assert backend_request.fresh == 72
    assert backend_request.cnfs == [expected_cnf]
    assert backend_request.ll_requests == [
        LowLevelRequest("EQ", 1, [23, 27, 31, 35]),
        LowLevelRequest("EQ", 1, [24, 28, 32, 36]),
        LowLevelRequest("EQ", 1, [25, 29, 33, 37]),
        LowLevelRequest("EQ", 1, [26, 30, 34, 38])
    ]


def test_fully_cross_with_uncrossed_simple_factors():
    other = Factor('other', ['l1', 'l2'])
    block = CrossBlock([color, text, other],
                       [color, text],
                       [])

    backend_request = BackendRequest(25)
    Cross.apply(block, backend_request)

    (expected_cnf, _) = to_cnf_tseitin(And([
        Iff(25, And([1,  3 ])), Iff(26, And([1,  4 ])), Iff(27, And([2,  3 ])), Iff(28, And([2,  4 ])),
        Iff(29, And([7,  9 ])), Iff(30, And([7,  10])), Iff(31, And([8,  9 ])), Iff(32, And([8,  10])),
        Iff(33, And([13, 15])), Iff(34, And([13, 16])), Iff(35, And([14, 15])), Iff(36, And([14, 16])),
        Iff(37, And([19, 21])), Iff(38, And([19, 22])), Iff(39, And([20, 21])), Iff(40, And([20, 22]))
    ]), 41)

    assert backend_request.fresh == 74
    assert backend_request.cnfs == [expected_cnf]
    assert backend_request.ll_requests == [
        LowLevelRequest("EQ", 1, [25, 29, 33, 37]),
        LowLevelRequest("EQ", 1, [26, 30, 34, 38]),
        LowLevelRequest("EQ", 1, [27, 31, 35, 39]),
        LowLevelRequest("EQ", 1, [28, 32, 36, 40])
    ]


def test_fully_cross_with_transition_in_crossing():
    direction = Factor("direction", ["up", "down"])

    block = CrossBlock([direction, color, color_repeats_factor],
                       [direction, color_repeats_factor],
                       [])

    backend_request = BackendRequest(29)
    Cross.apply(block, backend_request)

    (expected_cnf, _) = to_cnf_tseitin(And([
        Iff(29, And([ 5, 21])), Iff(30, And([ 5, 22])), Iff(31, And([ 6, 21])), Iff(32, And([ 6, 22])),
        Iff(33, And([ 9, 23])), Iff(34, And([ 9, 24])), Iff(35, And([10, 23])), Iff(36, And([10, 24])),
        Iff(37, And([13, 25])), Iff(38, And([13, 26])), Iff(39, And([14, 25])), Iff(40, And([14, 26])),
        Iff(41, And([17, 27])), Iff(42, And([17, 28])), Iff(43, And([18, 27])), Iff(44, And([18, 28])),
    ]), 45)

    assert backend_request.fresh == 78
    assert backend_request.cnfs == [expected_cnf]
    assert backend_request.ll_requests == [
        LowLevelRequest("EQ", 1, [29, 33, 37, 41]),
        LowLevelRequest("EQ", 1, [30, 34, 38, 42]),
        LowLevelRequest("EQ", 1, [31, 35, 39, 43]),
        LowLevelRequest("EQ", 1, [32, 36, 40, 44])
    ]


def test_fully_cross_with_exclude():
    color = Factor("color", ["red", "blue", "green"])
    text =  Factor("text",  ["red", "blue"])

    def illegal_stimulus(color, text):
        return color == "green" and text == "blue"

    def legal_stimulus(color, text):
        return not illegal_stimulus(color, text)

    stimulus_configuration = Factor("stimulus configuration", [
        DerivedLevel("legal",   WithinTrial(legal_stimulus, [color, text])),
        DerivedLevel("illegal", WithinTrial(illegal_stimulus, [color, text]))
    ])

    block = CrossBlock([color, text, stimulus_configuration],
                       [color, text],
                       [Exclude(stimulus_configuration["illegal"])],
                       require_complete_crossing=False)

    backend_request = BackendRequest(36)
    Cross.apply(block, backend_request)

    (expected_cnf, _) = to_cnf_tseitin(And([
        Iff(36, And([ 1, 4 ])), Iff(37, And([ 1, 5 ])), Iff(38, And([ 2, 4 ])), Iff(39, And([ 2, 5 ])), Iff(40, And([ 3, 4 ])),
        Iff(41, And([ 8, 11])), Iff(42, And([ 8, 12])), Iff(43, And([ 9, 11])), Iff(44, And([ 9, 12])), Iff(45, And([10, 11])),
        Iff(46, And([15, 18])), Iff(47, And([15, 19])), Iff(48, And([16, 18])), Iff(49, And([16, 19])), Iff(50, And([17, 18])),
        Iff(51, And([22, 25])), Iff(52, And([22, 26])), Iff(53, And([23, 25])), Iff(54, And([23, 26])), Iff(55, And([24, 25])),
        Iff(56, And([29, 32])), Iff(57, And([29, 33])), Iff(58, And([30, 32])), Iff(59, And([30, 33])), Iff(60, And([31, 32]))
    ]), 61)

    assert backend_request.fresh == 112
    assert backend_request.cnfs == [expected_cnf]
    assert backend_request.ll_requests == [
        LowLevelRequest("EQ", 1, [36, 41, 46, 51, 56]),
        LowLevelRequest("EQ", 1, [37, 42, 47, 52, 57]),
        LowLevelRequest("EQ", 1, [38, 43, 48, 53, 58]),
        LowLevelRequest("EQ", 1, [39, 44, 49, 54, 59]),
        LowLevelRequest("EQ", 1, [40, 45, 50, 55, 60])
    ]


def test_derivation():
    # Congruent derivation
    d = Derivation(4, [[0, 2], [1, 3]], con_factor)
    backend_request = BackendRequest(24)
    d.apply(block, backend_request)

    (expected_cnf, expected_fresh) = to_cnf_tseitin(And([
        Iff(5,  Or([And([1,  3 ]), And([2,  4 ])])),
        Iff(11, Or([And([7,  9 ]), And([8,  10])])),
        Iff(17, Or([And([13, 15]), And([14, 16])])),
        Iff(23, Or([And([19, 21]), And([20, 22])]))
    ]), 24)

    assert backend_request.fresh == expected_fresh
    assert backend_request.cnfs == [expected_cnf]

    # Incongruent derivation
    d = Derivation(5, [[0, 3], [1, 2]], con_factor)
    backend_request = BackendRequest(24)
    d.apply(block, backend_request)

    (expected_cnf, expected_fresh) = to_cnf_tseitin(And([
        Iff(6,  Or([And([1,  4 ]), And([2,  3 ])])),
        Iff(12, Or([And([7,  10]), And([8,  9 ])])),
        Iff(18, Or([And([13, 16]), And([14, 15])])),
        Iff(24, Or([And([19, 22]), And([20, 21])]))
    ]), 24)

    assert backend_request.fresh == expected_fresh
    assert backend_request.cnfs == [expected_cnf]


def test_derivation_with_transition():
    block = CrossBlock([color, text, color_repeats_factor],
                       [color, text],
                       [])

    # Color repeats derivation
    d = Derivation(16, [[0, 4], [1, 5]], color_repeats_factor)
    backend_request = BackendRequest(23)
    d.apply(block, backend_request)

    (expected_cnf, expected_fresh) = to_cnf_tseitin(And([
        Iff(17, Or([And([1, 5 ]), And([2,  6 ])])),
        Iff(19, Or([And([5, 9 ]), And([6,  10])])),
        Iff(21, Or([And([9, 13]), And([10, 14])]))
    ]), 23)

    assert backend_request.fresh == expected_fresh
    assert backend_request.cnfs == [expected_cnf]

    # Color does not repeat derivation
    d = Derivation(17, [[0, 5], [1, 4]], color_repeats_factor)
    backend_request = BackendRequest(23)
    d.apply(block, backend_request)

    (expected_cnf, expected_fresh) = to_cnf_tseitin(And([
        Iff(18, Or([And([1, 6 ]), And([2,  5 ])])),
        Iff(20, Or([And([5, 10]), And([6,  9 ])])),
        Iff(22, Or([And([9, 14]), And([10, 13])]))
    ]), 23)

    assert backend_request.fresh == expected_fresh
    assert backend_request.cnfs == [expected_cnf]


def test_derivation_with_multiple_transitions():
    block = CrossBlock([color, text, color_repeats_factor, text_repeats_factor],
                       [color, text],
                       [])

    # Text repeats derivation
    d = Derivation(22, [[2, 6], [3, 7]], text_repeats_factor)
    backend_request = BackendRequest(29)
    d.apply(block, backend_request)

    (expected_cnf, expected_fresh) = to_cnf_tseitin(And([
        Iff(23, Or([And([3,  7 ]), And([4,  8 ])])),
        Iff(25, Or([And([7,  11]), And([8,  12])])),
        Iff(27, Or([And([11, 15]), And([12, 16])]))
    ]), 29)

    assert backend_request.fresh == expected_fresh
    assert backend_request.cnfs == [expected_cnf]

    # Text does not repeat derivation
    d = Derivation(23, [[2, 7], [3, 6]], text_repeats_factor)
    backend_request = BackendRequest(29)
    d.apply(block, backend_request)

    (expected_cnf, expected_fresh) = to_cnf_tseitin(And([
        Iff(24, Or([And([3,  8 ]), And([4,  7 ])])),
        Iff(26, Or([And([7,  12]), And([8,  11])])),
        Iff(28, Or([And([11, 16]), And([12, 15])]))
    ]), 29)

    assert backend_request.fresh == expected_fresh
    assert backend_request.cnfs == [expected_cnf]


def test_derivation_with_three_level_transition():
    f = Factor("f", ["a", "b", "c"])
    f_transition = Factor("transition", [
        DerivedLevel("aa", Transition(lambda c: c[-1] == "a" and c[0] == "a", [f])),
        DerivedLevel("ab", Transition(lambda c: c[-1] == "a" and c[0] == "b", [f])),
        DerivedLevel("ac", Transition(lambda c: c[-1] == "a" and c[0] == "c", [f])),
        DerivedLevel("ba", Transition(lambda c: c[-1] == "b" and c[0] == "a", [f])),
        DerivedLevel("bb", Transition(lambda c: c[-1] == "b" and c[0] == "b", [f])),
        DerivedLevel("bc", Transition(lambda c: c[-1] == "b" and c[0] == "c", [f])),
        DerivedLevel("ca", Transition(lambda c: c[-1] == "c" and c[0] == "a", [f])),
        DerivedLevel("cb", Transition(lambda c: c[-1] == "c" and c[0] == "b", [f])),
        DerivedLevel("cc", Transition(lambda c: c[-1] == "c" and c[0] == "c", [f])),
    ])

    block = CrossBlock([f, f_transition], [f], [])

    # a-a derivation
    d = Derivation(9, [[0, 3]], f_transition)
    backend_request = BackendRequest(28)
    d.apply(block, backend_request)

    (expected_cnf, expected_fresh) = to_cnf_tseitin(And([
        Iff(10, Or([And([1, 4])])),
        Iff(19, Or([And([4, 7])]))
    ]), 28)

    assert backend_request.fresh == expected_fresh
    assert backend_request.cnfs == [expected_cnf]


def test_derivation_with_general_window():
    block = CrossBlock([color, text, congruent_bookend],
                       [color, text],
                       [])
    # congruent bookend - yes
    d = Derivation(16, [[0, 2], [1, 3]], congruent_bookend)
    backend_request = BackendRequest(19)
    d.apply(block, backend_request)

    (expected_cnf, expected_fresh) = to_cnf_tseitin(And([
        Iff(17, Or([And([ 1, 3 ]), And([ 2, 4 ])])),
        Iff(19, Or([And([13, 15]), And([14, 16])]))
    ]), 19)

    assert backend_request.fresh == expected_fresh
    assert backend_request.cnfs == [expected_cnf]

    # congruent bookend - no
    d = Derivation(17, [[0, 3], [1, 2]], congruent_bookend)
    backend_request = BackendRequest(19)
    d.apply(block, backend_request)

    (expected_cnf, expected_fresh) = to_cnf_tseitin(And([
        Iff(18, Or([And([ 1, 4 ]), And([ 2, 3 ])])),
        Iff(20, Or([And([13, 16]), And([14, 15])]))
    ]), 19)

    assert backend_request.fresh == expected_fresh
    assert backend_request.cnfs == [expected_cnf]


def test_atmostkinarow_validate():
    color = Factor("color", ["red", "blue"])
    with pytest.raises(ValueError):
        AtMostKInARow(SimpleLevel("yo"), color)

    # Levels must either be a factor/level tuple, or a Factor.
    AtMostKInARow(1, (color, "red"))
    AtMostKInARow(1, color)

    with pytest.raises(ValueError):
        AtMostKInARow(1, 42)

    with pytest.raises(ValueError):
        AtMostKInARow(1, (color, "red", "oops"))

    with pytest.raises(ValueError):
        AtMostKInARow(1, (color, "no such level"))


def __run_kinarow(c: Constraint, block: Block = block) -> BackendRequest:
    backend_request = BackendRequest(block.variables_per_sample() + 1)
    for dc in c.desugar({}):
        dc.apply(block, backend_request)
    return backend_request

def test_atleastkinarow():
    backend_request = __run_kinarow(AtLeastKInARow(2, (color, "red")))
    (expected_cnf, expected_fresh) = to_cnf_tseitin(And([
        If(1, And([7])),
        If(And([Not(1), 7]), And([13])),
        If(And([Not(7), 13]), And([19])),
        If(Not(13), Not(Or([19]))),
    ]), 25)

    assert backend_request.fresh == expected_fresh
    assert backend_request.cnfs == [expected_cnf]


    backend_request = __run_kinarow(AtLeastKInARow(3, (color, "red")))
    (expected_cnf, expected_fresh) = to_cnf_tseitin(And([
        If(1, And([7, 13])),
        If(And([Not(1), 7]), And([13, 19])),
        If(Not(7), Not(Or([13, 19]))),
    ]), 25)

    assert backend_request.fresh == expected_fresh
    assert backend_request.cnfs == [expected_cnf]

def test_atmostkinarow():
    backend_request = __run_kinarow(AtMostKInARow(3, color))
    assert backend_request.ll_requests == [
        LowLevelRequest("LT", 4, [1,  7, 13, 19]),
        LowLevelRequest("LT", 4, [2,  8, 14, 20])
    ]

    backend_request = __run_kinarow(AtMostKInARow(1, (color, "red")))
    assert backend_request.ll_requests == [
        LowLevelRequest("LT", 2, [1,  7 ]),
        LowLevelRequest("LT", 2, [7,  13]),
        LowLevelRequest("LT", 2, [13, 19])
    ]

    backend_request = __run_kinarow(AtMostKInARow(2, (color, "red")))
    assert backend_request.ll_requests == [
        LowLevelRequest("LT", 3, [1, 7,  13]),
        LowLevelRequest("LT", 3, [7, 13, 19])
    ]

    backend_request = __run_kinarow(AtMostKInARow(1, (color, "blue")))
    assert backend_request.ll_requests == [
        LowLevelRequest("LT", 2, [2,  8 ]),
        LowLevelRequest("LT", 2, [8,  14]),
        LowLevelRequest("LT", 2, [14, 20])
    ]

    backend_request = __run_kinarow(AtMostKInARow(2, (color, "blue")))
    assert backend_request.ll_requests == [
        LowLevelRequest("LT", 3, [2, 8,  14]),
        LowLevelRequest("LT", 3, [8, 14, 20])
    ]

    backend_request = __run_kinarow(AtMostKInARow(3, (con_factor, "con")))
    assert backend_request.ll_requests == [
        LowLevelRequest("LT", 4, [5, 11, 17, 23])
    ]


def test_atmostkinarow_disallows_k_of_zero():
    with pytest.raises(ValueError):
        AtMostKInARow(0, (con_factor, "con"))


def test_nomorethankinarow_sugar():
    backend_request = __run_kinarow(AtMostKInARow(1, (color, "red")))
    assert backend_request.ll_requests == [
        LowLevelRequest("LT", 2, [1,  7 ]),
        LowLevelRequest("LT", 2, [7,  13]),
        LowLevelRequest("LT", 2, [13, 19])
    ]


@pytest.mark.parametrize('design', permutations([color, text, color_repeats_factor]))
def test_atmostkinarow_with_transition(design):
    block = CrossBlock(design, [color, text], list(map(Reify, design)))

    backend_request = __run_kinarow(AtMostKInARow(1, (color_repeats_factor, "yes")), block)
    assert backend_request.ll_requests == [
        LowLevelRequest("LT", 2, [17, 19]),
        LowLevelRequest("LT", 2, [19, 21])
    ]

    backend_request = __run_kinarow(AtMostKInARow(1, (color_repeats_factor, "no")), block)
    assert backend_request.ll_requests == [
        LowLevelRequest("LT", 2, [18, 20]),
        LowLevelRequest("LT", 2, [20, 22])
    ]


def test_atmostkinarow_with_multiple_transitions():
    block = CrossBlock([color, text, color_repeats_factor, text_repeats_factor],
                       [color, text],
                       [Reify(color_repeats_factor), Reify(text_repeats_factor)])

    backend_request = __run_kinarow(AtMostKInARow(1, (text_repeats_factor, "yes")), block)
    assert backend_request.ll_requests == [
        LowLevelRequest("LT", 2, [23, 25]),
        LowLevelRequest("LT", 2, [25, 27])
    ]

    backend_request = __run_kinarow(AtMostKInARow(1, (text_repeats_factor, "no")), block)
    assert backend_request.ll_requests == [
        LowLevelRequest("LT", 2, [24, 26]),
        LowLevelRequest("LT", 2, [26, 28])
    ]


def test_exactlykinarow():
    backend_request = __run_kinarow(ExactlyKInARow(1, (color, "red")))
    (expected_cnf, expected_fresh) = to_cnf_tseitin(And([
        If(1, Not(7)),
        If(And([Not(1), 7]), Not(13)),
        If(And([Not(7), 13]), Not(19))
    ]), 25)

    assert backend_request.fresh == expected_fresh
    assert backend_request.cnfs == [expected_cnf]

    backend_request = __run_kinarow(ExactlyKInARow(2, (color, "red")))
    (expected_cnf, expected_fresh) = to_cnf_tseitin(And([
        If(1, And([7, Not(13)])),
        If(And([Not(1), 7]), And([13, Not(19)])),
        If(And([Not(7), 13]), 19),
        If(19, 13)
    ]), 25)

    assert backend_request.fresh == expected_fresh
    assert backend_request.cnfs == [expected_cnf]

    backend_request = __run_kinarow(ExactlyKInARow(3, (color, "red")))
    (expected_cnf, expected_fresh) = to_cnf_tseitin(And([
        If(1, And([7, 13, Not(19)])),
        If(And([Not(1), 7]), And([13, 19])),
        If(19, 13),
        If(13, 7)
    ]), 25)

    assert backend_request.fresh == expected_fresh
    assert backend_request.cnfs == [expected_cnf]


def test_exactlykinarow_disallows_k_of_zero():
    with pytest.raises(ValueError):
        ExactlyKInARow(0, (color, "red"))


def test_kinarow_with_bad_factor():
    bogus_factor = Factor("f", ["a", "b", "c"])
    with pytest.raises(ValueError):
        CrossBlock(design, crossing, [ExactlyKInARow(2, (bogus_factor, "a"))])




def test_exclude():
    f = Exclude(color["red"])
    backend_request = BackendRequest(0)
    f.apply(block, backend_request)
    assert backend_request.cnfs == [And([-1, -7, -13, -19])]

    f = Exclude(con_factor["con"])
    backend_request = BackendRequest(0)
    f.apply(block, backend_request)
    assert backend_request.cnfs == [And([-5, -11, -17, -23])]


def test_exclude_with_transition():
    block = CrossBlock([color, text, color_repeats_factor],
                       [color, text],
                       [])

    c = Exclude(color_repeats_factor["yes"])
    backend_request = BackendRequest(0)
    c.apply(block, backend_request)
    assert backend_request.cnfs == [And([-17, -19, -21])]


def test_exclude_with_general_window():
    block = CrossBlock([color, text, congruent_bookend],
                       [color, text],
                       [])

    c = Exclude(congruent_bookend["yes"])
    backend_request = BackendRequest(0)
    c.apply(block, backend_request)
    assert backend_request.cnfs == [And([-17, -19])]


def test_exclude_with_reduced_crossing():
    color = Factor("color", ["red", "blue", "green"])
    text =  Factor("text",  ["red", "blue"])

    def illegal_stimulus(color, text):
        return color == "green" and text == "blue"

    def legal_stimulus(color, text):
        return not illegal_stimulus(color, text)

    stimulus_configuration = Factor("stimulus configuration", [
        DerivedLevel("legal",   WithinTrial(legal_stimulus, [color, text])),
        DerivedLevel("illegal", WithinTrial(illegal_stimulus, [color, text]))
    ])

    c = Exclude(stimulus_configuration["illegal"])
    block = CrossBlock([color, text, stimulus_configuration],
                       [color, text],
                       [c],
                       require_complete_crossing=False)

    backend_request = BackendRequest(0)
    c.apply(block, backend_request)
    assert backend_request.cnfs == [And([-7, -14, -21, -28, -35])]


def test_exclude_with_three_derived_levels():
    color_list = ["red", "green", "blue"]
    color = Factor("color", color_list)
    text  = Factor("text",  color_list)

    def count_diff(colors, texts):
        changes = 0
        if (colors[0] != colors[-1]): changes += 1
        if (texts[0] != texts[-1]): changes += 1
        return changes

    def make_k_diff_level(k):
        def k_diff(colors, texts):
            return count_diff(colors, texts) == k
        return DerivedLevel(str(k), Transition(k_diff, [color, text]))

    changed = Factor("changed", [make_k_diff_level(0),
                                 make_k_diff_level(1),
                                 make_k_diff_level(2)]);

    exclude_constraint = Exclude(changed["2"])

    design       = [color, text, changed]
    crossing     = [color, text]
    block        = CrossBlock(design, crossing, [exclude_constraint])

    backend_request = BackendRequest(0)
    exclude_constraint.apply(block, backend_request)
    assert backend_request.cnfs == [And([-57, -60, -63, -66, -69, -72, -75, -78])]

def test_pin():
    f = Pin(0, color["red"])
    backend_request = BackendRequest(0)
    f.apply(block, backend_request)
    assert backend_request.cnfs == [And([1])]

    f = Pin(-1, con_factor["con"])
    backend_request = BackendRequest(0)
    f.apply(block, backend_request)
    assert backend_request.cnfs == [And([23])]

    f = Pin(1000, color["red"])
    backend_request = BackendRequest(0)
    f.apply(block, backend_request)
    assert backend_request.cnfs == [And([1, -1])]



from sweetpea._internal.sampling_strategy.base import Gen, SamplingResult
from sweetpea._internal.block import Block
from sweetpea._internal.sampling_strategy.unigen import UniGen
from sweetpea._internal.sampling_strategy.random import RandomGen

"""
This represents a uniform sampling strategy where a specific strategy
is selected automatically. When requsting more than one sample, however,
the selected strategy may or may not sample with repetition.
"""
class UniformGen(Gen):
    
    @staticmethod
    def class_name():
        return 'UniformGen'

    @staticmethod
    def sample(block: Block, sample_count: int) -> SamplingResult:
        if block.complex_factors_or_constraints:
            return UniGen.sample(block, sample_count)
        else:
            return RandomGen.sample(block, sample_count)



import glob
import operator as op
import os
import pytest
import re

from sweetpea import CrossBlock, Repeat, MinimumTrials, synthesize_trials, UniformGen
from sweetpea._internal.primitive import Factor, DerivedLevel, WithinTrial, Transition, Window
from sweetpea._internal.constraint import Exclude, ExactlyKInARow, AtMostKInARow, Reify
from sweetpea._internal.sampling_strategy.random import RandomGen, UCSolutionEnumerator

color = Factor("color", ["red", "blue"])
text  = Factor("text",  ["red", "blue"])

red_color = color["red"]
blue_color = color["blue"]
red_text = text["red"]
blue_text = text["blue"]

con_factor_within_trial = Factor("congruent?", [
    DerivedLevel("con", WithinTrial(op.eq, [color, text])),
    DerivedLevel("inc", WithinTrial(op.ne, [color, text]))
])

con_factor_window = Factor("congruent?", [
    DerivedLevel("con", Window(op.eq, [color, text], 1, 1)),
    DerivedLevel("inc", Window(op.ne, [color, text], 1, 1))
])

color_repeats_factor = Factor("repeated color?", [
    DerivedLevel("yes", Transition(lambda colors: colors[0] == colors[-1], [color])),
    DerivedLevel("no",  Transition(lambda colors: colors[0] != colors[-1], [color]))
])


def test_validate_accepts_basic_factors():
    block = CrossBlock([color, text],
                       [color, text],
                       [])
    RandomGen._RandomGen__validate(block)


def test_validate_accepts_derived_factors_with_simple_windows():
    block = CrossBlock([color, text, con_factor_within_trial],
                       [color, text],
                       [])
    RandomGen._RandomGen__validate(block)

    block = CrossBlock([color, text, con_factor_window],
                       [color, text],
                       [])
    RandomGen._RandomGen__validate(block)


def test_validate_accepts_implied_derived_factors_with_complex_windows():
    block = CrossBlock([color, text, color_repeats_factor],
                       [color, text],
                       [])
    # no error

def test_example_counts():
    # Get all the python examples from the uc-counting-tests directory
    path_to_test_files = os.path.dirname(os.path.abspath(__file__)) + "/uc-counting-tests/*.py"
    files = glob.glob(path_to_test_files)

    failures = []

    for filename in files:
        contents = None
        with open(filename, 'r') as f:
            contents = f.read()
            exec(contents, globals(), locals())

        if 'block' not in vars():
            failures.append("File did not produce a variable named 'block', aborting. file={}".format(filename))
            continue

        matches = re.search('# ASSERT COUNT = (\d+)', contents)
        if matches:
            expected_count = int(matches.groups(0)[0])
            enumerator = UCSolutionEnumerator(vars()['block'])
            if enumerator.solution_count() != expected_count:
                failures.append("Count Mismatch. Actual count: {}, Expected count: {}, File={}"
                    .format(enumerator.solution_count(), expected_count, filename))
        else:
            failures.append("File did not contain an assertion for count, aborting. file={}".format(filename))

    if failures:
        pytest.fail('{} failures occurred in counting tests: {}'.format(len(failures), failures))


def test_constraint_violation():
    are_constraints_violated = RandomGen._RandomGen__are_constraints_violated

    block = CrossBlock([color, text, con_factor_within_trial],
                       [color, text],
                       [ExactlyKInARow(2, (color, red_color))])

    class FakeEnumer:
        def __init__(self):
            self.has_crossed_complex_derived_factors = False
    enumer = FakeEnumer()

    assert are_constraints_violated(block, {color: [color[l] for l in ['red', 'red', 'blue', 'blue']]}, enumer, 4, 0, 0) == False
    assert are_constraints_violated(block, {color: [color[l] for l in ['red', 'blue', 'red', 'blue']]}, enumer, 4, 0, 0) == True

    block = CrossBlock([color, text, con_factor_within_trial],
                       [color, text],
                       [AtMostKInARow(2, (color, red_color))])

    assert are_constraints_violated(block, {color: [color[l] for l in ['red', 'blue', 'blue', 'blue']]}, enumer, 4, 0, 0) == False
    assert are_constraints_violated(block, {color: [color[l] for l in ['red', 'red', 'blue', 'blue']]}, enumer, 4, 0, 0) == False
    assert are_constraints_violated(block, {color: [color[l] for l in ['red', 'red', 'red', 'blue']]}, enumer, 4, 0, 0) == True
    assert are_constraints_violated(block, {color: [color[l] for l in ['blue', 'red', 'red', 'red']]}, enumer, 4, 0, 0) == True

def test_minimum_trials_repeat():
    for min_trials in [1, 2, 3, 4, 5, 6, 7, 17, 55]:
        block = Repeat(CrossBlock([color, text],
                                  [color, text],
                                  []),
                       [MinimumTrials(min_trials)])
        runs = synthesize_trials(block=block, samples=1, sampling_strategy=RandomGen)
        assert len(runs[0]["color"]) == max(4, min_trials)

def test_minimum_trials():
    for min_trials in [1, 2, 3, 4, 5, 6, 7]:
        block = CrossBlock([color, text],
                           [color, text],
                           [MinimumTrials(min_trials)])
        runs = synthesize_trials(block=block, samples=1, sampling_strategy=RandomGen)
        assert len(runs[0]["color"]) == max(4, min_trials)



"""This module provides functionality for handling logic formulas."""


from collections import namedtuple
from functools import reduce
from itertools import product
from typing import Any, Dict, List, Tuple, Union, cast


And = namedtuple('And', 'input_list')
Or  = namedtuple('Or',  'input_list')
If  = namedtuple('If',  'p q')
Iff = namedtuple('Iff', 'p q')
Not = namedtuple('Not', 'c')

Formula = Union[And, Or, Not, int]
FormulaWithIff = Union[And, Or, If, Iff, Not, int]
FormulaAndFresh = Tuple[Formula, int]


# Simple Cache class used by the Tseitin transformation. Maintains the next fresh variable as
# state, along with the cached values.
class _Cache:
    def __init__(self, next_variable: int) -> None:
        self.cache = cast(Dict[str, int], {})
        self.next_variable = next_variable

    def get(self, s: str) -> int:
        if s in self.cache:
            return self.cache[s]
        else:
            self.cache[s] = self.next_variable
            self.next_variable += 1
            return self.cache[s]

    def get_next_variable(self) -> int:
        return self.next_variable


def to_cnf_naive(f: FormulaWithIff, next_variable: int) -> Tuple[And, int]:
    """Converts to CNF using the pattern described on
    https://en.wikipedia.org/wiki/Conjunctive_normal_form#Converting_from_first-order_logic.

    The conversion is naive, which means it is subject to an exponential
    explosion in the length of the formula. The upside is that it does not
    introduce any new variables, which constrains the overall size of the
    solution space. (#SAT)
    """
    formula = __eliminate_iff(f)
    formula = __apply_demorgan(formula)
    formula = __distribute_ors_naive(formula)
    if not isinstance(formula, And):
        formula = And([formula])
    return (formula, next_variable)


def to_cnf_switching(f: FormulaWithIff, next_variable: int) -> Tuple[And, int]:
    """Converts to CNF using the pattern described on
    https://en.wikipedia.org/wiki/Conjunctive_normal_form#Converting_from_first-order_logic.

    However, distributing ORs over ANDs is done via switching variables, to
    keep the converted formula small, as described here at
    https://www.cs.jhu.edu/~jason/tutorials/convert-to-CNF.html.
    """
    formula = __eliminate_iff(f)
    formula = __apply_demorgan(formula)
    (formula, fresh) = __distribute_ors_switching(formula, next_variable)
    if not isinstance(formula, And):
        formula = And([formula])
    return (formula, fresh)


def to_cnf_tseitin(f: FormulaWithIff, next_variable: int) -> Tuple[And, int]:
    """Converts to CNF using the Tseitin transformation. This will introduce
    addtional variables, which will increase #SAT, but provides a linear bound
    on the length increase of the new formula. See
    https://en.wikipedia.org/wiki/Tseytin_transformation.
    """
    clauses = cast(List[Formula], [])
    cache = _Cache(next_variable)

    new_rep = __tseitin_rep(f, clauses, cache)
    clauses.append(new_rep)

    return (And(clauses), cache.get_next_variable())


def cnf_to_json(formula: List[And]) -> List[List[int]]:
    or_list = []
    for a in formula:
        for o in a.input_list:
            if isinstance(o, Or):
                l = cast(List[int], [])
                for n in o.input_list:
                    if isinstance(n, int):
                        l.append(n)
                    elif isinstance(n, Not):
                        l.append(-n.c)
                    else:
                        raise ValueError("Value was not Not tuple or int")
                or_list.append(l)
            elif isinstance(o, int):
                or_list.append([o])
            else:
                raise ValueError("Value was not Or tuple or variable!")
    return or_list


def __eliminate_iff(f: FormulaWithIff) -> Formula:
    if isinstance(f, And):
        return And(list(map(__eliminate_iff, f.input_list)))
    elif isinstance(f, Or):
        return Or(list(map(__eliminate_iff, f.input_list)))
    elif isinstance(f, If):
        return __eliminate_iff(Or([Not(f.p), f.q]))
    elif isinstance(f, Iff):
        new_formula = And([
            Or([f.p, Not(f.q)]),
            Or([Not(f.p), f.q])
        ])
        return __eliminate_iff(new_formula)
    elif isinstance(f, Not):
        return Not(__eliminate_iff(f.c))
    elif isinstance(f, int):
        return f


def __apply_demorgan(f: Formula) -> Formula:
    if isinstance(f, And):
        return __build_and(list(map(__apply_demorgan, f.input_list)))
    elif isinstance(f, Or):
        return __build_or(list(map(__apply_demorgan, f.input_list)))
    elif isinstance(f, Not):
        clause = cast(Formula, f.c)
        if isinstance(clause, And):
            return __apply_demorgan(__build_or(list(map(lambda c: Not(c), clause.input_list))))
        elif isinstance(clause, Or):
            return __apply_demorgan(__build_and(list(map(lambda c: Not(c), clause.input_list))))
        elif isinstance(clause, Not):
            return __apply_demorgan(clause.c)
        elif isinstance(clause, int):
            return f
    elif isinstance(f, int):
        return f


def __distribute_ors_naive(f: Formula) -> Formula:
    if isinstance(f, And):
        return __build_and(list(map(__distribute_ors_naive, f.input_list)))
    elif isinstance(f, Or):
        clauses = list(map(__distribute_ors_naive, f.input_list))
        crossable_clauses = list(map(__get_list_for_crossing, clauses))
        crossed_clauses = list(product(*crossable_clauses))
        or_list = list(map(__build_or, crossed_clauses))
        return __build_and(cast(List[Formula], or_list))
    elif isinstance(f, Not) or isinstance(f, int):
        return f


def __distribute_ors_switching(f: Formula, fresh: int) -> FormulaAndFresh:
    if isinstance(f, And):
        (clauses, new_fresh) = reduce(__apply_distribute_ors,
                                      f.input_list,
                                      (cast(List[Formula], []), fresh))
        return (__build_and(clauses), new_fresh)
    elif isinstance(f, Or):
        (clauses, new_fresh) = reduce(__apply_distribute_ors,
                                      f.input_list,
                                      (cast(List[Formula], []), fresh))
        clauses.sort(key=__order_clauses)
        if len(clauses) > 1:
            if __should_not_combine(clauses):
                return (f, fresh)
            elif __should_combine_naively(clauses):
                return __distribute_ors_switching(__naive_combination(clauses), new_fresh)
            else:
                (new_formula, new_fresh) = __switching_combination(clauses, new_fresh)
                return __distribute_ors_switching(new_formula, new_fresh)
        else:
            return (clauses[0], new_fresh)
    elif isinstance(f, Not):
        assert isinstance(f.c, int)
        return (f, fresh)
    elif isinstance(f, int):
        return (f, fresh)


def __flatten_clause_list(clauses: List[Formula], cls: Any) -> List[Formula]:
    flattened_list = cast(List[Formula], [])
    for c in clauses:
        if isinstance(c, cls):
            flattened_list.extend(c.input_list)
        else:
            flattened_list.append(c)
    flattened_list.sort(key=__order_clauses)
    return flattened_list


def __build_or(l: List[Formula]) -> Or:
    return Or(__flatten_clause_list(l, Or))


def __build_and(l: List[Formula]) -> And:
    return And(__flatten_clause_list(l, And))


def __apply_distribute_ors(acc: Tuple[List[Formula], int], elem: Formula) -> Tuple[List[Formula], int]:
    fs = acc[0]
    fresh = acc[1]
    (f, new_fresh) = __distribute_ors_switching(elem, fresh)
    fs.append(f)
    return (fs, new_fresh)


def __should_not_combine(clauses: List[Formula]) -> bool:
    return not any(isinstance(c, And) for c in clauses)


def __should_combine_naively(clauses: List[Formula]) -> bool:
    return \
        isinstance(clauses[0], int) or \
        isinstance(clauses[0], Not) or \
        isinstance(clauses[1], int) or \
        isinstance(clauses[1], Not)


def __naive_combination(clauses: List[Formula]) -> Formula:
    lhs = __get_list_for_crossing(clauses[0])
    rhs = __get_list_for_crossing(clauses[1])
    crossing = [list(tup) for tup in list(product(lhs, rhs))]
    combination = cast(Formula, And(list(map(lambda l: Or(__flatten_clause_list(l, Or)), crossing))))
    if len(clauses) > 2:
        return __build_or([combination] + clauses[2:])
    else:
        return combination


def __get_list_for_crossing(clause: Formula) -> List[Formula]:
    if isinstance(clause, int) or isinstance(clause, Not):
        return [clause]
    elif isinstance(clause, And) or isinstance(clause, Or):
        return clause.input_list


def __switching_combination(clauses: List[Formula], fresh: int) -> FormulaAndFresh:
    lhs = Or([Not(fresh), clauses[0]])
    rhs = Or([fresh, clauses[1]])
    combination = cast(Formula, And([lhs, rhs]))
    new_fresh = fresh + 1
    if len(clauses) > 2:
        return (__build_or([combination] + clauses[2:]), new_fresh)
    else:
        return (combination, new_fresh)


def __order_clauses(c: Formula) -> int:
    if isinstance(c, And) or isinstance(c, Or):
        return 0
    elif isinstance(c, Not):
        return c.c
    else:
        return c


def __tseitin_rep(f: FormulaWithIff,
                  clauses: List[Formula],
                  cache: _Cache) -> Formula:
    if isinstance(f, And):
        # Replace any subformulae
        new_vars = list(map(lambda c: __tseitin_rep(c, clauses, cache), f.input_list))

        # Get the variable that represents this clause
        old_next_var = cache.get_next_variable()
        new_rep = cache.get(str(And(new_vars)))

        # Record the equivalences, if the cache missed.
        if old_next_var == new_rep:
            clauses.append(Or(cast(List[Formula], list(map(Not, new_vars))) +
                              cast(List[Formula], [new_rep])))
            clauses.extend(list(map(lambda v: Or([v, Not(new_rep)]), new_vars)))

        return new_rep

    elif isinstance(f, Or):
        # Replace any subformulae
        new_vars = list(map(lambda c: __tseitin_rep(c, clauses, cache), f.input_list))

        # Get the variable that represents this clause
        old_next_var = cache.get_next_variable()
        new_rep = cache.get(str(Or(new_vars)))

        # Record the equivalences, if the cache missed.
        if old_next_var == new_rep:
            clauses.append(Or(cast(List[Formula], new_vars) +
                              cast(List[Formula], [Not(new_rep)])))
            clauses.extend(list(map(lambda v: Or([Not(v), new_rep]), new_vars)))

        return new_rep

    elif isinstance(f, If):
        # Replace any subformulae
        new_p = __tseitin_rep(f.p, clauses, cache)
        new_q = __tseitin_rep(f.q, clauses, cache)

         # Get the variable that represents this clause
        old_next_var = cache.get_next_variable()
        new_rep = cache.get(str(If(new_p, new_q)))

        # Record the equivalences, if the cache missed.
        if old_next_var == new_rep:
            clauses.append(Or([Not(new_p), new_q, Not(new_rep)]))
            clauses.append(Or([    new_p,  new_rep]))
            clauses.append(Or([Not(new_q), new_rep]))

        return new_rep

    elif isinstance(f, Iff):
        # Replace any subformulae
        new_p = __tseitin_rep(f.p, clauses, cache)
        new_q = __tseitin_rep(f.q, clauses, cache)

        # Get the variable that represents this clause
        old_next_var = cache.get_next_variable()
        new_rep = cache.get(str(Iff(new_p, new_q)))

        # Record the equivalences, if the cache missed.
        if old_next_var == new_rep:
            clauses.append(Or([    new_p,      new_q,      new_rep ]))
            clauses.append(Or([Not(new_p), Not(new_q),     new_rep ]))
            clauses.append(Or([    new_p,  Not(new_q), Not(new_rep)]))
            clauses.append(Or([Not(new_p),     new_q,  Not(new_rep)]))

        return new_rep

    elif isinstance(f, Not):
        # Replace any subformulae in the Not clause.
        new_f = __tseitin_rep(f.c, clauses, cache)

        # Allocate a new variable to represent the new Not clause
        old_next_var = cache.get_next_variable()
        new_rep = cache.get(str(Not(new_f)))

        # Record the equivalence between the new representation and the original.
        if old_next_var == new_rep:
            clauses.append(Or([    new_f,      new_rep ]))
            clauses.append(Or([Not(new_f), Not(new_rep)]))

        # Return it.
        return new_rep

    elif isinstance(f, int):
        return f
